#!/bin/bash
#
#  Copyright FUJITSU LIMITED 2013
#
#  INSTALL.sh - install/update script
#
export LANG=C
PATH=/bin:/usr/bin:/usr/sbin:$PATH

# get shell path/name
CURDIR=`pwd`
cd `dirname $0`
SH_PATH=`pwd`
cd $CURDIR
SH_NAME=`basename $0`

# Install Definition
RPMSPATH="RPMS"
SOFTNAME=`basename $SH_PATH`
PKGS=`find $SH_PATH/$RPMSPATH -name "*.rpm" -print 2>/dev/null`
PKGS=`rpm -qp --qf '%{NAME}\n' $PKGS`

# Environment Definition
STDOUT="/dev/stdout"

# LOG file
MSGLOG="/tmp/FJSVdr-util-setup.log"
MSGID="FJSVdr-util"

#  Install time
DATE=`date '+%b %d %Y'`

# Initialize
RESULT=0
F_INSTALL=0
F_UPDATE=0
F_FORCE=0
S_INSTALL=0
S_ALREADY=0
ARCH=`uname -i`

#
# Output message function
#   usage: msg msg-text
#
msg()
{
	echo "$MSGID : $*" | tee -a $MSGLOG
}

#
# Shell exit function
#   usage: shell_exit [code]
#
shell_exit()
{
	if [ $F_UPDATE = 0 ]; then
		ACTION="install"
	else
		ACTION="update"
	fi
	if [ $1 != 0 ]; then
		msg "$SOFTNAME $ACTION failed."
		ret=1
	else
		msg "$SOFTNAME $ACTION completed."
		ret=0
	fi

	echo "" >> $MSGLOG
	exit $ret
}

do_rpm()
{
	msg "#" rpm "$@"
	rpm "$@" 2>&1 | tee -a $MSGLOG
	ret=${PIPESTATUS[0]}
	return $ret
} 

#
# install function
#   usage: install rpm-file ...
#   result code: =0(completed), !=0(failed)
#
install()
{
	local rpm="$1"

	# set rpm options
	OPT=""
	if [ $F_FORCE != 0 ]; then
		OPT="$OPT --force --nodeps"
	fi
	# execute rpm
	do_rpm -ivh $OPT "$rpm"
	ret=$?
	if [ $ret -eq 0 ]; then
		return 0
	fi
	return $ret
}

starts_with()
{
    local name="$1"
    local pat="$2"

    test "${name#$pat}" != "$name"
}

base_starts_with()
{
    local name="$1"
    local pat="$2"

    starts_with "$(basename "$name")" "$pat"
}

rpm_name()
{
    local name="$1"

    rpm -q --queryformat="%{NAME}.%{ARCH}\n" -p "$name"
}

#
# update function
#   usage: update rpm-file ...
#   result code: =0(completed), !=0(failed)
#
update()
{
	local rpm="$1"
	local rpmname=$(rpm_name "$rpm")

	# set rpm options
	OPT=""
	if [ $F_FORCE != 0 ]; then
		OPT="$OPT --force --nodeps"
	fi
	# execute rpm
	if base_starts_with "$rpm" "kmod-" ; then
	    do_rpm -e $OPT "$rpmname"
	    do_rpm -ivh $OPT "$rpm"
	fi
	ret=$?
	if [ $ret -eq 0 ]; then
		return 0
	fi
	return $ret
}

#
# return the name of rpmfile from the package name.
#   usage: package_to_rpmfile package
#   result code: =0(completed), !=0(failed)
#
package_to_rpmfile()
{
	local rpm
	local pkg="$1"
	local pkgname

	for rpm in `find $SH_PATH/$RPMSPATH -name "*.rpm" | grep -v 'src\.rpm'`; do

		pkgname=`rpm -qp --qf '%{NAME}' "$rpm" 2>/dev/null`
		if [ "$pkgname" == "$pkg" ]; then
			echo $rpm
			return 0
		fi
	done

	return 1
}

#
# usage
#   usage: usage
#
usage()
{
	msg "Usage: $SH_NAME [--install | --update] [--force]"
}

# NAME    : is_supported_machine
# USAGE   : is_supported_machine
function is_supported_machine()
{
    local DMIDECODE
    local model
    local ret="false"

    if [ "$ARCH" = "x86_64" ]; then

        # Fujitsu standard method to identify the machine type.
        DMIDECODE=/usr/sbin/dmidecode model=$(${DMIDECODE} -t 1 2>/dev/null | awk -F: '$1 ~ /Product Name$/ { print substr($2,match($2,/[[:alpha:]]/)); exit 0}')
        if [ "$model" ]; then
            echo $model | egrep -i 'PRIMEQUEST' &>/dev/null
            if [ $? -eq 0 ]; then
                # The machine is PRIMEQUEST
                ret="true"
            fi
        else
            # dmidecode failed. Try another non-standard way.
            lshal | grep 'system.hardware.product' | egrep 'PRIMEQUEST' &>/dev/null
            if [ $? -eq 0 ]; then
                # The machine is PRIMEQUEST
                ret="true"
            fi
        fi
    fi

    echo $ret
}

#
# Main routine
#

# user account check (root)
if [ $EUID != 0 ]; then
	msg "$SH_NAME: only root can do that."
	shell_exit 1
fi

# check argument
for ARG in $*; do
	case $ARG in
	"--install")
		F_INSTALL=1
		;;
	"--update")
		F_UPDATE=1
		;;
	"--force")
		F_FORCE=1
		;;
	*)
		msg "$SH_NAME: option $ARG invalid."
		usage
		shell_exit 1
		;;
	esac
done

echo "$MSGID : FJSVdr-util install start at $DATE" >> $MSGLOG

if [ $F_INSTALL != 0 -a $F_UPDATE != 0 ]; then
	msg "$SH_NAME: cannot specify --install with --update."
	usage
	shell_exit 1
fi

if [ $F_UPDATE != 0 -o $F_INSTALL != 0 ]; then
	F_FORCE=0
fi

# machine check
SUPPORTED=`is_supported_machine`
if [ "$SUPPORTED" = "false" ]; then
        msg "$SH_NAME: not supported machine."
        shell_exit 1
fi

# package loop
for PKG in $PKGS; do

	# initialization
	S_ALREADY=0
	S_INSTALL=0

	# check/get package file
	RPMFILE=`package_to_rpmfile $PKG`
	if [ "$RPMFILE" = "" ]; then
		msg "$SOFTNAME package not found."
		shell_exit 1
	fi
	NEWPKG=`basename $RPMFILE | sed -n "s/\.[^\.]*\.rpm$//p" 2>/dev/null`

	# check/get installed package
	NOWPKG=`rpm -q $PKG 2>/dev/null` && S_INSTALL=1

	# check obsoletes package is installed
	PKGNAME=`find $SH_PATH/$RPMSPATH -name "$PKG*.rpm" -print 2>/dev/null`
	OBSPKG=`rpm -qp --obsoletes $PKGNAME 2>/dev/null`
	NOWPKG=`rpm -q $OBSPKG 2>/dev/null` && S_INSTALL=1

	# check already installed new package
	if ([ $S_INSTALL != 0 ] && [ $F_FORCE = 0 ]); then
		RPMMSG="`rpm -Uvh --test $RPMFILE 2>&1`" && RPMMSG=""
		echo $RPMMSG | grep "already installed" >/dev/null && S_ALREADY=1
	fi

	# package update (if new package is already installed)
	if [ $S_ALREADY != 0 ]; then
		msg "new package $NOWPKG is already installed."

	# package update (if old package is already installed)
	elif [ $S_INSTALL != 0 ]; then
		if [ $F_INSTALL = 0 ]; then
			F_UPDATE=1
			update $RPMFILE
			if [ $? = 0 ]; then
				msg "package $NEWPKG is updated."
			else
				msg "package $NEWPKG is update failed."
				RESULT=1
			fi
		else
			msg "new package $NOWPKG is already installed."
		fi
	# package install
	else
		install $RPMFILE
		if [ $? = 0 ]; then
			msg "package $NEWPKG is installed."
		else
			msg "package $NEWPKG is install failed."
			RESULT=1
		fi
	fi

done

# exit (and output message)
shell_exit $RESULT

