#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# library for dr command
#
# Copyright(C) 2012 MUNEDA Takahiro
# Copyright(C) 2012 FUJITSU LIMITED
#

import os
import re
import sys
import time
import errno
import socket
import syslog
import inspect
import subprocess

DP_CMD_VERSION = '1.5.0'
DP_CMD_HOME = '/opt/FJSVdr-util'
DP_CORE_PATH = "%s/libexec/dp-core" % DP_CMD_HOME
DP_USER_COMMAND_PATH = "%s/user_command" % DP_CMD_HOME
DP_CONF_PATH = "%s/etc" % DP_CMD_HOME
DP_LOG_PATH = "%s/var/log" % DP_CMD_HOME
DP_PID_FILE = '/var/run/dr.pid'
DP_SYSFS = '/sys'
DP_PROCFS = '/proc'
DP_SB_MAX = 4  # Up to this number of SBs are available
DP_IOU_MAX = 4  # Up to this number of IOUs are available
DP_SB_REMOVE_RETRY = 5
DP_SB_REMOVE_RETRY_WAIT = 30  # Wait these seconds before going to retry
DP_DEBUG = 0

POWERON = 1
POWEROFF = 0
EMPTY = 2
DP_COLLAB_TIMINGS = ['ADD_PRE', 'ADD_POST', 'RM_PRE', 'RM_POST']

OS_RELEASE_FILE = "/etc/os-release"

pci_addr_regexp = (
    re.compile(r'^[0-9a-f]{4}:[0-9a-f]{2}:[0-9a-f]{2}\.[0-9a-f]$'))


class SB(object):
    class Node(object):
        def __init__(self, node):
            '''
            An object to manage node
            .node(int): node number
            .cpus(str): list of CPU nums on this node
            .mems(str): list of MEM nums on this node
            .succeeded_cpus(str): list of CPU nums succeeded to hotplug
            .succeeded_mems(str): list of MEM nums succeeded to hotplug
            .failed_cpus(str): list of CPU nums failed to hotplug
            .failed_mems(str): list of MEM nums failed to hotplug
            @node(int): node number
            '''
            self.node = node
            self.cpus = []
            self.mems = []
            self.succeeded_cpus = []
            self.succeeded_mems = []
            self.failed_cpus = []
            self.failed_mems = []

        def ScanCPUs(self):
            '''
            Scan CPUs under the Node and add them to the Node.cpus list
            Return value: None
            '''
            cpu_entry_regexp = re.compile(r'^cpu([\d]+)$')
            node_path = "%s/devices/system/node/node%s" % (DP_SYSFS, self.node)
            if not os.path.isdir(node_path):
                LogDebug("%s: not a directory" % node_path)
                return
            for entry in os.listdir(node_path):
                if not cpu_entry_regexp.match(entry):
                    continue
                cpu_num = cpu_entry_regexp.sub(r'\1', entry)
                self.cpus.append(cpu_num)
            if self.cpus:
                self.cpus.sort(lambda x, y: cmp(int(x), int(y)))

        def ScanMEMs(self):
            '''
            Scan Memories under the Node and add them to the Node.mems list
            Return value: None
            '''
            mem_entry_regexp = re.compile(r'^memory([\d]+)$')
            node_path = "%s/devices/system/node/node%s" % (DP_SYSFS, self.node)
            if not os.path.isdir(node_path):
                LogDebug("%s: not a directory" % node_path)
                return
            for entry in os.listdir(node_path):
                if not mem_entry_regexp.match(entry):
                    continue
                mem_num = mem_entry_regexp.sub(r'\1', entry)
                self.mems.append(mem_num)
            if self.mems:
                self.mems.sort(lambda x, y: cmp(int(x), int(y)))

    def __init__(self, dev):
        '''
        An object to manage SB.
        .dev(int): SB number
        .nodes(str): list of node nums on SB
        .node_objs(node object): list of node objects
        .devpath(str): sysfs device path for this SB (RHEL6.4 compatibility)
        @dev(int): SB number
        '''
        self.dev = dev
        self.nodes = []
        self.node_objs = []
        self.devpath = None

    def HasNonMovableZone(self):
        '''
        Check whether zones other than Movable zone are on the SB
        Return value(int): Has non movable zone or error -> 1, not -> 0
        '''
        zone_regexp = re.compile(r'^Node ([\d]+),\s+zone\s+([\w]+)\n$')

        path = "%s/zoneinfo" % DP_PROCFS
        lines, error = ReadlinesFile(path)
        if error:
            LogDebug("Can not read %s/zoneinfo with error %d" %
                     (DP_PROCFS, error))
            return 1
        if not lines:
            LogDebug("%s/zoneinfo is empty" % DP_PROCFS)
            return 1

        for line in lines:
            if 'Node' in line:
                node_num = zone_regexp.sub(r'\1', line)
                zone_type = zone_regexp.sub(r'\2', line)
            else:
                pass
            if not node_num in self.nodes:
                continue
            if 'present' in line:
                zone_allocated = int(line.strip().strip()[-1])
                if not zone_type == 'Movable' and zone_allocated != 0:
                    LogDebug("%s: Node%s has non Movable zone, maybe more" %
                             (zone_type, node_num))
                    return 1
        return 0

    def _FindSB(self):
        '''
        Find an ACPI0004:NN directory for the SB
        Return value(str): name of directory or None if not exists
        '''
        if self.devpath:
            # RHEL6.4 compatibility
            entry = self.devpath.split('/')[-1]
            if 'ACPI0004:' in entry:
                return entry
            else:
                return None

        path = "%s/bus/acpi/devices" % DP_SYSFS
        if not os.path.isdir(path):
            LogDebug("%s: not a directory" % path)
            return None
        for entry in os.listdir(path):
            if not 'ACPI0004:' in entry:
                continue
            # On RHEL7.1 or later, ACPI0004 directory always exists.
            # To check ACPI0004 is poweron or not, the status file must
            # be checked on RHEL7.1 or later
            sta_path = "%s/%s/status" % (path, entry)
            if os.path.exists(sta_path):
                sta, error = ReadFile(sta_path)
                if sta is None:
                    return None
                if int(sta) == 0:
                    continue
            sun_path = "%s/%s/sun" % (path, entry)
            sun, error = ReadFile(sun_path)
            if str(self.dev) == sun:
                return entry
        return None

    def _FindNodeswNodeSymlinks(self):
        '''
        Find NUMA nodes with node symlink files
        Return value: None
        Note: those symlinks are available RHEL6.5 and later
        '''
        # Node number can find by following path:
        # RHEL7.0 or later:
        #   /sys/bus/acpi/devices/ACPI0004:NN/LNXCPU:MM/physical_node/nodeX/
        # RHEL7.1 or later for Haswall:
        #   /sys/bus/acpi/devices/ACPI0004:NN/ACPI0007:MM/physical_node/nodeX/
        # RHEL6 (6.5 or later):
        #  /sys/bus/acpi/devices/ACPI0004:NN/LNXCPU:MM/sysdev/nodeX/
        node_regexp = re.compile(r'node([\d]+)$')

        acpi_path = self._FindSB()
        if not acpi_path:
            return

        sb_path = "%s/bus/acpi/devices/%s" % (DP_SYSFS, acpi_path)
        if not os.path.isdir(sb_path):
            LogDebug("%s: not a directory" % sb_path)
            return
        for entry in os.listdir(sb_path):
            if not 'LNXCPU' in entry and not 'ACPI0007' in entry:
                continue
            # try new sysfs layout first
            path = "%s/%s/physical_node" % (sb_path, entry)
            if not os.path.isdir(path):
                LogDebug("%s: not a directory" % path)
                # Hmm, maybe running on RHEL6
                path = "%s/%s/sysdev" % (sb_path, entry)
                if not os.path.isdir(path):
                    LogDebug("%s: not a directory" % path)
                    continue
            for node_entry in os.listdir(path):
                if not node_regexp.match(node_entry):
                    continue
                node_num = node_regexp.sub(r'\1', node_entry)
                if node_num in self.nodes:
                    continue
                self.nodes.append(node_num)

    def _FindNodeswoNodeSymlinks(self):
        '''
        Find NUMA nodes without node symlink files
        Return value: None
        Note: this is a backward compatibility method for RHEL6.4
        '''
        # Node number can find by checking the CPUs on SB and nodes
        # - RHEL6 or RHEL7
        #   /sys/bus/acpi/devices/ACPI0004:NN/LNXCPU:MM/sysdev ->
        #      ../../../../system/cpu/cpuX
        # - RHEL7.1 or later for Haswell
        #   /sys/bus/acpi/devices/ACPI0004:NN/ACPI0007:MM/sysdev ->
        #      ../../../../system/cpu/cpuX
        # /sys/devices/system/node/nodeY/cpuX -> ../../cpu/cpuX
        # nodeY is on this SB(ACPI0004:NN)
        cpu_regexp = re.compile(r'cpu\d+$')
        node_regexp = re.compile(r'node(\d+)$')

        acpi_path = self._FindSB()
        if not acpi_path:
            return

        sb_path = "%s/bus/acpi/devices/%s" % (DP_SYSFS, acpi_path)
        if not os.path.isdir(sb_path):
            LogDebug("%s: not a directory" % sb_path)
            return

        # find all cpus on this sb
        acpi_cpus = []
        for entry in os.listdir(sb_path):
            if not 'LNXCPU' in entry and not 'ACPI0007' in entry:
                continue
            path = "%s/%s/sysdev" % (sb_path, entry)
            if not os.path.exists(path):
                continue
            acpi_cpu = os.path.realpath(path).split('/')[-1]
            if not cpu_regexp.match(acpi_cpu):
                continue
            acpi_cpus.append(acpi_cpu)
        if not acpi_cpus:
            return

        # find all nodes on this system
        node_path = "%s/devices/system/node" % DP_SYSFS
        if not os.path.isdir(node_path):
            return
        nodes = []
        for entry in os.listdir(node_path):
            if not node_regexp.match(entry):
                continue
            node_num = node_regexp.sub(r'\1', entry)
            nodes.append(node_num)
        if not nodes:
            return

        # find all nodes these cpus are on
        for node in nodes:
            for cpu in acpi_cpus:
                if os.path.exists("%s/devices/system/node/node%s/%s" %
                                  (DP_SYSFS, node, cpu)):
                    self.nodes.append(node)
                    break

    def _FindNodes(self):
        '''
        Find Numa nodes under the SB, add node numbers to the SB.nodes
        list, and create node object for them.
        Return value: None
        '''
        # try new sysfs layout first
        self._FindNodeswNodeSymlinks()
        if not self.nodes:  # error or old sysfs layout
            self._FindNodeswoNodeSymlinks()

        if self.nodes:
            self.nodes.sort(lambda x, y: cmp(int(x), int(y)))
            for node_num in self.nodes:
                node_obj = self.Node(int(node_num))
                self.node_objs.append(node_obj)

    def _ShowCurrent(self):
        '''
        Show amount of resources currently available on this SB
        Return value(int): success -> 0, error -> 1
        '''
        cpu_lists = []  # list of contents of .../node/nodeX/cpulist
        cpu_nums = []   # list of cpu numbers
        mems = []       # list of MemTotal line of .../node/nodeX/meminfo
        mem_nums = []   # list of memory entry under .../node/nodeX/
        mem_entry_regexp = re.compile(r'^memory([\d]+)$')

        retval = 0
        self._FindNodes()
        for node_num in self.nodes:
            node_path = "%s/devices/system/node/node%s" % (DP_SYSFS, node_num)
            if not os.path.isdir(node_path):
                LogDebug("%s: not a directory" % node_path)
                retval += 1
                continue

            cpulist_path = "%s/cpulist" % node_path
            cpu_list, error = ReadFile(cpulist_path)
            if cpu_list is None:
                retval += 1
            else:
                cpu_lists.append(cpu_list)

            meminfo_path = "%s/meminfo" % node_path
            meminfo, error = ReadFile(meminfo_path)
            if meminfo is None:
                retval += 1
            else:
                for mem in meminfo.splitlines():
                    if 'MemTotal' in mem:
                        mem_amount = re.sub(r'.*MemTotal:\s*([\d]+) kB', r'\1',
                                            mem)
                        mems.append(mem_amount)
                        break

            for entry in os.listdir(node_path):
                if not mem_entry_regexp.match(entry):
                    continue
                mem_num = mem_entry_regexp.sub(r'\1', entry)
                mem_path = ("%s/devices/system/memory/memory%s/state" %
                            (DP_SYSFS, mem_num))
                mem_status, error = ReadFile(mem_path)
                if mem_status is None:
                    retval += 1
                    continue
                elif not mem_status == 'online':
                    continue
                else:
                    mem_nums.append(mem_num)

        # example of cpu_lists: ['0-3,15-18', '10-15,20']
        # To reorder like following, extract abbrev., reorder and
        # convert to abbrev. -> ['0-3,10-18,20']
        for cpu_list in cpu_lists:
            cpu_nums += AbbrevToList(cpu_list)
        cpu_abbrev = ListToAbbrev(cpu_nums)
        sys.stdout.write("CPU: %s: %d\n" % (cpu_abbrev[0], cpu_abbrev[1]))
        mem_total = 0
        for mem_amount in mems:
            mem_total += int(mem_amount)
        mem_abbrev = ListToAbbrev(mem_nums)
        sys.stdout.write("MEM: %s: %d kiB\n" % (mem_abbrev[0], mem_total))
        node_abbrev = ListToAbbrev(self.nodes)
        sys.stdout.write("Node: %s\n" % node_abbrev[0])
        return retval

    def _ShowPossible(self):
        '''
        Show amount of all resources including currently disabled on this SB
        Return value(int): Always 0
        '''
        cpu_nums = []  # list of cpu entry of under .../node/nodeX/
        cpu_entry_regexp = re.compile(r'^cpu([\d]+)$')
        mem_nums = []  # list of memory entry of under .../node/nodeX/
        mem_entry_regexp = re.compile(r'^memory([\d]+)$')

        self._FindNodes()
        for node_num in self.nodes:
            node_path = "%s/devices/system/node/node%s" % (DP_SYSFS, node_num)
            if not os.path.isdir(node_path):
                LogDebug("%s: not a directory" % node_path)
                continue
            for entry in os.listdir(node_path):
                if cpu_entry_regexp.match(entry):
                    cpu_num = cpu_entry_regexp.sub(r'\1', entry)
                    cpu_nums.append(cpu_num)
                elif mem_entry_regexp.match(entry):
                    mem_num = mem_entry_regexp.sub(r'\1', entry)
                    mem_nums.append(mem_num)

        cpu_abbrev = ListToAbbrev(cpu_nums)
        sys.stdout.write("CPU: %s: %d\n" % (cpu_abbrev[0], cpu_abbrev[1]))
        mem_abbrev = ListToAbbrev(mem_nums)
        sys.stdout.write("MEM: %s\n" % mem_abbrev[0])
        node_abbrev = ListToAbbrev(self.nodes)
        sys.stdout.write("Node: %s\n" % node_abbrev[0])
        return 0

    def Show(self, possible=0):
        '''
        Show amount of resources on this SB
        @possible(int): Show all resources including currently disabled.
        Return value(int): success -> 0, error -> 1
        '''
        if not possible:
            return self._ShowCurrent()
        else:
            return self._ShowPossible()

    def Eject(self):
        '''
        Write 0 to /path/to/ACPI0004:NN/physical_node/online then
        Write 1 to /path/to/ACPI0004:NN/eject to eject SB
        Return value(int): success -> 0, error -> errno
        '''
        acpi_path = self._FindSB()
        power_path = "%s/bus/acpi/devices/%s/physical_node/online" % (DP_SYSFS, acpi_path)
        eject_path = "%s/bus/acpi/devices/%s/eject" % (DP_SYSFS, acpi_path)

        if os.path.isfile(power_path):
            retval = WriteFile(power_path, '0')
            if retval:
                LogDebug("%s: failed to write" % power_path)
                LogErr("424 : Failed to eject SB%d with error %d" % (self.dev, retval))
                return retval

        retval = WriteFile(eject_path, '1')
        if retval == 0:
            sta_path = "%s/bus/acpi/devices/%s/status" % (DP_SYSFS, acpi_path)
            acpi_dir = "%s/bus/acpi/devices/%s" % (DP_SYSFS, acpi_path)
            elapsed_seconds = 0
            timeout = 3600 # wait for 60m
            while elapsed_seconds < timeout:
                # On RHEL7.1 or later, ACPI0004 directory always exists.
                # To check ACPI0004 is poweron or not, the status file must
                # be checked on RHEL7.1 or later
                if os.path.exists(sta_path):
                    sta, error = ReadFile(sta_path)
                    if sta is None:
                        continue
                    if int(sta) == 0:
                        return 0
                else:
                    if not os.path.exists(acpi_dir):
                        return 0
                time.sleep(1)
                elapsed_seconds += 1

            LogDebug("%s: cannot not delete " % acpi_dir)
            retval = errno.EBUSY

        LogErr("424 : Failed to eject SB%d with error %d" % (self.dev, retval))
        return retval

    def _Status(self, dev, dev_type):
        '''
        Check cpu/mem status
        @dev(str): Target device number
        @devtype(str): 'cpu' or 'memory'
        Return value(int): offline -> 0, online -> 1, error -> -errno
        '''
        if not isinstance(dev, str) or not re.match(r'[0-9]+$', dev):
            if isinstance(dev, tuple):
                dev = (dev,)
            LogDebug("%s: wrong dev was passed" % dev)
            return -errno.EINVAL
        if not dev_type == 'cpu' and not dev_type == 'memory':
            if isinstance(dev_type, tuple):
                dev_type = (dev_type,)
            LogDebug("%s: wrong dev_type was passed" % dev_type)
            return -errno.EINVAL

        if dev_type == 'cpu':
            power_path = "%s/devices/system/cpu/cpu%s/online" % (DP_SYSFS, dev)
        else:
            power_path = ("%s/devices/system/memory/memory%s/state" %
                          (DP_SYSFS, dev))

        status = 0
        if dev_type == 'cpu':
            cpu_status, error = ReadFile(power_path)
            if cpu_status is None:
                return -error
            status = int(cpu_status)
        else:
            mem_status, error = ReadFile(power_path)
            # offline or online
            if mem_status is None:
                return -error
            elif mem_status == 'online':
                status = POWERON
            else:
                status = POWEROFF
        return status

    def _PowerControl(self, power, dev, dev_type):
        '''
        Power Off/On the CPU/MEM
        @power(int): 0 -> PowerOff, 1 -> PowerOn.
        @dev(str): Target device number
        @dev_type(str): 'cpu' or 'memory'
        Return value(int): success -> 0, error -> errno
        '''
        if not isinstance(power, int) or not 0 <= power <= 1:
            if isinstance(power, tuple):
                power = (power,)
            LogDebug("%s: wrong power was passed" % power)
            return errno.EINVAL
        if not isinstance(dev, str) or not re.match(r'[0-9]+$', dev):
            if isinstance(dev, tuple):
                dev = (dev,)
            LogDebug("%s: wrong dev was passed" % dev)
            return errno.EINVAL
        if not dev_type == 'cpu' and not dev_type == 'memory':
            if isinstance(dev_type, tuple):
                dev_type = (dev_type,)
            LogDebug("%s: wrong dev_type was passed" % dev_type)
            return errno.EINVAL

        if dev_type == 'cpu':
            power_path = "%s/devices/system/cpu/cpu%s/online" % (DP_SYSFS, dev)
        else:
            power_path = ("%s/devices/system/memory/memory%s/state" %
                          (DP_SYSFS, dev))

        if dev_type == 'memory':
            if power == POWEROFF:
                target = 'offline'
            else:
                major = CheckDistroVersion()[0]
                if int(major) >= 12:
                    target = 'online_movable'
                else:
                    target = 'online'
        else:
            target = str(power)

        retval = WriteFile(power_path, target)
        return retval

    def PowerOn(self, dev, dev_type):
        '''
        Power On a device and make sure it goes well
        @dev(str): Target device number
        @dev_type(str): 'cpu' or 'memory'
        Return value(int): success -> 0, error -> 1
        '''
        if not isinstance(dev, str) or not re.match(r'[0-9]+$', dev):
            if isinstance(dev, tuple):
                dev = (dev,)
            LogDebug("%s: wrong dev was passed" % dev)
            LogErr('470 : Internal Error')
            return 1
        if not dev_type == 'cpu' and not dev_type == 'memory':
            if isinstance(dev_type, tuple):
                dev_type = (dev_type,)
            LogDebug("%s: wrong dev_type was passed" % dev_type)
            LogErr('471 : Internal Error')
            return 1

        # Check power before controlling
        status = self._Status(dev, dev_type)
        if status == POWEROFF:
            pass
        elif status == POWERON:
            LogErr("280 : %s%s: device is already available" %
                   (dev_type, dev))
            return 0
        else:
            LogDebug("%s%s: error at before power controlling: %s" %
                     (dev_type, dev, status))
            LogErr('472 : Internal Error')
            return 1

        # Power controlling
        retval = self._PowerControl(POWERON, dev, dev_type)
        if retval == 0:
            pass
        else:
            if dev_type == 'cpu':
                msg_id = '400'
            else:
                msg_id = '402'
            LogErr("%s : Failed to add %s%s with error: %s" %
                   (msg_id, dev_type, dev, retval))
            return 1

        # Confirm it's really powered on
        status = self._Status(dev, dev_type)
        if status == POWERON:
            return 0
        else:
            if status == POWEROFF:
                if dev_type == 'cpu':
                    msg_id = '406'
                else:
                    msg_id = '407'
                LogErr("%s : %s%s status is invalid: offline" %
                       (msg_id, dev_type, dev))
            else:
                LogDebug("%s%s: error at after power controlling: %s" %
                         (dev_type, dev, status))
                LogErr('473 : Internal Error')
            return 1

    def CPUPowerOn(self, dev):
        '''
        Wrapper method for CPU power on
        @dev(str): CPU number
        Return value: return value from self.PowerOn()
        '''
        return self.PowerOn(dev, 'cpu')

    def MEMPowerOn(self, dev):
        '''
        Wrapper method for MEM power on
        @dev(str): MEM number
        Return value: return value from self.PowerOn()
        '''
        return self.PowerOn(dev, 'memory')

    def PowerOff(self, dev, dev_type):
        '''
        Power On a device and make sure it goes well
        @dev(str): Target device number
        @dev_type(str): 'cpu' or 'memory'
        Return value(int): success -> 0, error -> 1
        '''
        if not isinstance(dev, str) or not re.match(r'[0-9]+$', dev):
            if isinstance(dev, tuple):
                dev = (dev,)
            LogDebug("%s: wrong dev was passed" % dev)
            LogErr('474 : Internal Error')
            return 1
        if not dev_type == 'cpu' and not dev_type == 'memory':
            if isinstance(dev_type, tuple):
                dev_type = (dev_type,)
            LogDebug("%s: wrong dev_type was passed" % dev_type)
            LogErr('475 : Internal Error')
            return 1

        # Check power before controlling
        status = self._Status(dev, dev_type)
        if status == POWERON:
            pass
        elif status == POWEROFF:
            LogErr("287 : %s%s: device is already disabled" %
                   (dev_type, dev))
            return 0
        else:
            LogDebug("%s%s: error at before power controlling: %s" %
                     (dev_type, dev, status))
            LogErr('476 : Internal Error')
            return 1

        # Power controlling
        retval = self._PowerControl(POWEROFF, dev, dev_type)
        if retval == 0:
            pass
        else:
            if dev_type == 'cpu':
                msg_id = '420'
            else:
                msg_id = '422'
            LogErr("%s : Failed to remove %s%s with error: %s" %
                   (msg_id, dev_type, dev, retval))
            return 1

        # Confirm it's really powered off
        status = self._Status(dev, dev_type)
        if status == POWEROFF:
            return 0
        else:
            if status == POWERON:
                if dev_type == 'cpu':
                    msg_id = '408'
                else:
                    msg_id = '409'
                LogErr("%s : %s%s status is invalid: online" %
                       (msg_id, dev_type, dev))
            else:
                LogDebug("%s%s: error at after power controlling: %s" %
                         (dev_type, dev, status))
                LogErr('477 : Internal Error')
            return 1

    def CPUPowerOff(self, dev):
        '''
        Wrapper method for CPU power off
        @dev(str): CPU number
        Return value: return value from self.PowerOff()
        '''
        return self.PowerOff(dev, 'cpu')

    def MEMPowerOff(self, dev):
        '''
        Wrapper method for MEM power off
        @dev(str): MEM number
        Return value: return value from self.PowerOff()
        '''
        return self.PowerOff(dev, 'memory')

    def _SetupArg(self, timing, result=''):
        '''
        Set up arguments which are passed to collaboration programs
        @timing(str): Which event this method is called with.
        @result(str): Hotplug result, SUCCESS or FAILURE.  This is set
        when timing is *_POST.
        Return value(str): list of arguments (for subprocess) or None
        '''
        if not isinstance(timing, str) or not timing in DP_COLLAB_TIMINGS:
            if isinstance(timing, tuple):
                timing = (timing,)
            LogDebug("%s: wrong timing was passed" % timing)
            return []
        if 'POST' in timing:
            if not result or not isinstance(result, str):
                if isinstance(result, tuple):
                    result = (result,)
                LogDebug("%s: wrong result was passed" % result)
                return []
            elif not result == 'SUCCESS' and not result == 'FAILURE':
                LogDebug("%s: wrong result was passed" % result)
                return []

        args = ['-p', "SB%d" % self.dev]
        args += ['-e', "%s" % timing]

        if 'POST' in timing:
            args += ['-r', "%s" % result]
        else:  # ADD_PRE or RM_PRE
            cpus = []
            mems = []
            for node_obj in self.node_objs:
                cpus += node_obj.cpus
                mems += node_obj.mems
            allcpu = ListToAbbrev(cpus)
            allmem = ListToAbbrev(mems)
            if not allmem[0]:
                mem_amount = '0'
            else:
                # 128 MiB per one mem entry
                mem_amount = allmem[1] * 128 * 1024
            nodes = ListToAbbrev(self.nodes)
            args += ['-c', "%s" % allcpu[0], '-m', "%s" % mem_amount,
                     '-n', "%s" % nodes[0]]

        LogDebug("args: %s" % (' ').join(args))
        return args

    def Add(self):
        '''
        Online all CPU/MEM devices on the SB
        Return value(int):
          success -> 0x0
          error -> 0x2 : pre collab call
                   0x3 : cpu/mem addition
                   0x7 : post collab call
                   0xa : cpu/mem addition and post collab call
        '''
        # Find Nodes, CPUs, MEMs under the SB
        self._FindNodes()
        for node_obj in self.node_objs:
            node_obj.ScanCPUs()
            node_obj.ScanMEMs()

        node_abbrev = ListToAbbrev(self.nodes)
        LogInfo("801 : Added SB%d, Node%s" % (self.dev, node_abbrev[0]))

        collab_ret = 0
        collab_list = _FindCollabs()
        if collab_list:
            collab_arg = self._SetupArg(timing='ADD_PRE')
            collab_ret = CallCollabs("sb%d" % self.dev, 'ADD_PRE', collab_arg,
                                     collab_list)
            if collab_ret:
                return 0x2

        cpu_lists = []
        mem_lists = []
        for node_obj in self.node_objs:
            cpu_lists += node_obj.cpus
            mem_lists += node_obj.mems
        cpu_abbrev = ListToAbbrev(cpu_lists)
        LogInfo("802 : Add CPU%s (total %s)" % (cpu_abbrev[0], cpu_abbrev[1]))
        mem_abbrev = ListToAbbrev(mem_lists)
        mem_total = mem_abbrev[1] * 128 * 1024  # 128 MiB per one mem entry
        LogInfo("804 : Add MEM%s (total %d kiB)" % (mem_abbrev[0], mem_total))

        success = 0
        for node_obj in self.node_objs:
            # Memory device need to turn on descending order
            mem_list = node_obj.mems
            mem_list.reverse()
            for mem in mem_list:
                retval = self.MEMPowerOn(mem)
                if retval == 0:
                    node_obj.succeeded_mems.append(mem)
                else:
                    node_obj.failed_mems.append(mem)
                success += retval
            for cpu in node_obj.cpus:
                retval = self.CPUPowerOn(cpu)
                if retval == 0:
                    node_obj.succeeded_cpus.append(cpu)
                else:
                    node_obj.failed_cpus.append(cpu)
                success += retval

        succeeded_cpus = []
        succeeded_mems = []
        for node_obj in self.node_objs:
            if node_obj.succeeded_cpus:
                succeeded_cpus += node_obj.succeeded_cpus
            if node_obj.succeeded_mems:
                succeeded_mems += node_obj.succeeded_mems
        if succeeded_cpus:
            cpu_abbrev = ListToAbbrev(succeeded_cpus)
            LogInfo("803 : Added CPU%s (total %s)" %
                    (cpu_abbrev[0], cpu_abbrev[1]))
        if succeeded_mems:
            mem_abbrev = ListToAbbrev(succeeded_mems)
            mem_total = mem_abbrev[1] * 128 * 1024  # 128 MiB per one mem entry
            LogInfo("805 : Added MEM%s (total %d kiB)" %
                    (mem_abbrev[0], mem_total))

        retval = 0
        if success == 0:
            result = 'SUCCESS'
            LogInfo("809 : Added SB%d" % self.dev)
        else:
            result = 'FAILURE'
            failed_cpus = []
            failed_mems = []
            for node_obj in self.node_objs:
                if node_obj.failed_cpus:
                    failed_cpus += node_obj.failed_cpus
                if node_obj.failed_mems:
                    failed_mems += node_obj.failed_mems

            if failed_cpus:
                cpu_abbrev = ListToAbbrev(failed_cpus)
                LogErr("401 : Failed to add CPU%s (total %s)" %
                       (cpu_abbrev[0], cpu_abbrev[1]))
            if failed_mems:
                mem_abbrev = ListToAbbrev(failed_mems)
                # 128 MiB per one mem entry
                mem_total = mem_abbrev[1] * 128 * 1024
                LogErr("403 : Failed to add MEM%s (total %d kiB)" %
                       (mem_abbrev[0], mem_total))
            LogErr("405 : Failed to add SB%d" % self.dev)

        if collab_list:
            collab_arg = self._SetupArg(timing='ADD_POST', result=result)
            collab_ret = CallCollabs("sb%d" % self.dev, 'ADD_POST', collab_arg,
                                     collab_list)
            if collab_ret:
                if result == 'SUCCESS':
                    return 0x7
                else:
                    return 0xa

        if result == 'SUCCESS':
            return 0x0
        else:
            return 0x3

    def OfflineDevsOnSB(self):
        '''
        Offline CPU/MEM devices on node
        Return value(int): some of them have failed -> 1, or not -> 0
        '''
        if not self.node_objs:
            return 0

        success = 0
        for node_obj in self.node_objs:
            # Initialize succeeded/failure devices list for this run
            node_obj.succeeded_cpus = []
            node_obj.failed_cpus = []
            node_obj.succeeded_mems = []
            node_obj.failed_mems = []

            for cpu in node_obj.cpus:
                retval = self.CPUPowerOff(cpu)
                if retval == 0:
                    node_obj.succeeded_cpus.append(cpu)
                else:
                    node_obj.failed_cpus.append(cpu)
                success += retval
            for mem in node_obj.mems:
                retval = self.MEMPowerOff(mem)
                if retval == 0:
                    node_obj.succeeded_mems.append(mem)
                else:
                    node_obj.failed_mems.append(mem)
                success += retval

        succeeded_cpus = []
        succeeded_mems = []
        for node_obj in self.node_objs:
            if node_obj.succeeded_cpus:
                succeeded_cpus += node_obj.succeeded_cpus
            if node_obj.succeeded_mems:
                succeeded_mems += node_obj.succeeded_mems
        if succeeded_cpus:
            cpu_abbrev = ListToAbbrev(succeeded_cpus)
            LogInfo("823 : Removed CPU%s (total %s)" %
                    (cpu_abbrev[0], cpu_abbrev[1]))
        if succeeded_mems:
            mem_abbrev = ListToAbbrev(succeeded_mems)
            mem_total = mem_abbrev[1] * 128 * 1024  # 128 MiB per one mem entry
            LogInfo("825 : Removed MEM%s (total %d kiB)" %
                    (mem_abbrev[0], mem_total))

        if success == 0:
            return 0
        else:
            return 1

    def Remove(self):
        '''
        Offline all CPU/MEM devices on the SB.  If made them, then eject SB
        Return value(int):
          success -> 0x0
          error -> 0x1 : has non movable zone
                   0x2 : pre collab call
                   0x4 : cpu/mem removal (eject is not invoked)
                   0x7 : post collab call
                   0xb : cpu/mem removal and post collab call
                   0xe : cpu/mem removal (eject is invoked)
        '''
        # Find Nodes, CPUs, MEMs under the SB
        self._FindNodes()
        for node_obj in self.node_objs:
            node_obj.ScanCPUs()
            node_obj.ScanMEMs()

        if self.HasNonMovableZone():
            LogErr("230 : SB%d: invalid device is set" % self.dev)
            return 0x1

        node_abbrev = ListToAbbrev(self.nodes)
        LogInfo("821 : Remove SB%d, Node%s" % (self.dev, node_abbrev[0]))

        collab_ret = 0
        collab_list = _FindCollabs()
        if collab_list:
            collab_arg = self._SetupArg(timing='RM_PRE')
            collab_ret = CallCollabs("sb%d" % self.dev, 'RM_PRE', collab_arg,
                                     collab_list)
            if collab_ret:
                return 0x2

        cpu_lists = []
        mem_lists = []
        for node_obj in self.node_objs:
            cpu_lists += node_obj.cpus
            mem_lists += node_obj.mems
        cpu_abbrev = ListToAbbrev(cpu_lists)
        LogInfo("822 : Remove CPU%s (total %s)" %
                (cpu_abbrev[0], cpu_abbrev[1]))
        mem_abbrev = ListToAbbrev(mem_lists)
        mem_total = mem_abbrev[1] * 128 * 1024  # 128 MiB per one mem entry
        LogInfo("824 : Remove MEM%s (total %d kiB)" %
                (mem_abbrev[0], mem_total))

        success = sb_remove_retry = 0
        while sb_remove_retry < DP_SB_REMOVE_RETRY:
            success = self.OfflineDevsOnSB()
            if success == 0:
                # all offline operation succeeded
                break

            # prepare for retry
            sb_remove_retry += 1
            for node_obj in self.node_objs:
                # move failed devs list to the available devs list
                node_obj.cpus = node_obj.failed_cpus
                node_obj.mems = node_obj.failed_mems
            LogWarning('428 : Some CPUs and/or MEMs are busy, will retry')
            time.sleep(DP_SB_REMOVE_RETRY_WAIT)

        eject_invoked = 'false'
        if success == 0:
            success += self.Eject()
            eject_invoked = 'true'


        if success == 0:
            result = 'SUCCESS'
            LogInfo("829 : Removed SB%d" % self.dev)
        else:
            result = 'FAILURE'
            failed_cpus = []
            failed_mems = []
            for node_obj in self.node_objs:
                if node_obj.failed_cpus:
                    failed_cpus += node_obj.failed_cpus
                if node_obj.failed_mems:
                    failed_mems += node_obj.failed_mems

            if failed_cpus:
                cpu_abbrev = ListToAbbrev(failed_cpus)
                LogErr("421 : Failed to remove CPU%s (total %s)" %
                       (cpu_abbrev[0], cpu_abbrev[1]))
            if failed_mems:
                mem_abbrev = ListToAbbrev(failed_mems)
                # 128 MiB per one mem entry
                mem_total = len(failed_mems) * 128 * 1024
                LogErr("423 : Failed to remove MEM%s (total %d kiB)" %
                       (mem_abbrev[0], mem_total))
            LogErr("426 : Failed to remove SB%d" % self.dev)

        if collab_list:
            collab_arg = self._SetupArg(timing='RM_POST', result=result)
            collab_ret = CallCollabs("sb%d" % self.dev, 'RM_POST', collab_arg,
                                     collab_list)
            if collab_ret:
                if result == 'SUCCESS':
                    return 0x7
                else:
                    return 0xb

        if result == 'SUCCESS':
            return 0x0
        else:
            if eject_invoked == 'false':
                return 0x4
            else:
                LogDebug("Eject failed")
                return 0xe


class PciHotPlugSlot(object):
    def __init__(self, dev):
        '''
        An object to manage PciHotPlugSlot
        .slot_num(str): PCIe hotplug slot number
        .slot_addr(str): PCI address(seg:bus:dev)
        @dev(int): PCIe hotplug slot number
        '''
        self.slot_num = str(dev)
        self.slot_addr = ''

    def _GetSlotAddr(self):
        '''
        Get PCI address (seg:bus:dev) of this hotplug slot
        Return value(str): PCI address(seg:bus:dev)
        '''
        path = "%s/bus/pci/slots/%s/address" % (DP_SYSFS, self.slot_num)
        address, error = ReadFile(path)
        return address

    def ScanPciDevs(self, pci_addr):
        '''
        Scan PCI devices under the pci_addr
        @pci_addr(str): PCI address(seg:bus:dev.fn)
        Return value(str): sorted list of PCI address(seg:bus:dev.fn)
        '''
        if not isinstance(pci_addr, str):
            if isinstance(pci_addr, tuple):
                pci_addr = (pci_addr,)
            LogDebug("%s: wrong pci_addr was passed" % pci_addr)
            return []
        if not pci_addr_regexp.match(pci_addr):
            LogDebug("%s: wrong pci_addr was passed" % pci_addr)
            return []

        pci_devs = []
        path = "%s/bus/pci/devices/%s" % (DP_SYSFS, pci_addr)
        if not os.path.isdir(path):
            LogDebug("%s: not a directory" % path)
            return []
        pci_devs.append(pci_addr)
        for root, dirs, files in os.walk(path):
            for entry in dirs:
                if pci_addr_regexp.match(entry):
                    pci_devs.append(entry)

        pci_devs.sort()
        return pci_devs

    def _ScanPciDevsOnSlot(self):
        '''
        Get list of PCI devices on this PCI slot.
        Return value(str): sorted list of PCI addresses (seg:bus:dev.fn)
        '''
        if not self.slot_addr:
            return []

        pci_devs = []
        path = "%s/bus/pci/devices" % DP_SYSFS
        if not os.path.isdir(path):
            LogDebug("%s: not a directory" % path)
            return []
        for entry in os.listdir(path):
            if self.slot_addr in entry:
                pci_devs.append(entry)

        pci_devs.sort()
        return pci_devs

    def Show(self, show_slot_num=0):
        '''
        Show PCI/PCIe device(s) info below this PCIe slot
        @show_slot_num(int): if set, show slot number at line head
        Return value(int): 0
        '''
        if not isinstance(show_slot_num, int):
            if isinstance(show_slot_num, tuple):
                show_slot_num = (show_slot_num,)
            LogDebug("%s: wrong show_slot_num was passed" % show_slot_num)
            return 0

        self.slot_addr = self._GetSlotAddr()
        if not self.slot_addr:
            return 0

        pci_devs = []
        pcidev_list = self._ScanPciDevsOnSlot()
        for pci_dev in pcidev_list:
            pci_devs += self.ScanPciDevs(pci_dev)

        for pci_dev in pci_devs:
            output = CallLspciS(pci_dev)
            if show_slot_num:
                msg = "pcie%s: %s" % (self.slot_num, output)
            else:
                msg = output
            sys.stdout.write("%s\n" % msg)
        return 0

    def _Status(self):
        '''
        Check slot status
        Return value(int): offline -> 0, online -> 1, empty -> 2,
        error -> -errno
        '''
        power_path = "%s/bus/pci/slots/%s/power" % (DP_SYSFS, self.slot_num)
        power, error = ReadFile(power_path)
        if power is None:
            return -error
        elif power == str(POWERON):
            return 1

        adapter_path = ("%s/bus/pci/slots/%s/adapter" %
                        (DP_SYSFS, self.slot_num))
        adapter, error = ReadFile(adapter_path)
        if adapter is None:
            return -error
        elif adapter == str(POWERON):
            return 0  # offline
        else:
            return 2  # empty

    def _PowerControl(self, power):
        '''
        Power Off/On the PCIe slot
        @power(int): 0 -> PowerOff, 1 -> PowerOn.
        Return value(int): success -> 0, error -> errno
        '''
        if not isinstance(power, int) or not 0 <= power <= 1:
            if isinstance(power, tuple):
                power = (power,)
            LogDebug("%s: wrong power was passed" % power)
            return errno.EINVAL

        power_path = "%s/bus/pci/slots/%s/power" % (DP_SYSFS, self.slot_num)
        retval = WriteFile(power_path, str(power))
        return retval

    def PowerOn(self):
        '''
        Wrapper method for PCIe device(include IOU) power on
        Return value: return value from self._PowerControl()
        '''
        return self._PowerControl(POWERON)

    def PowerOff(self):
        '''
        Wrapper method for PCIe device(include IOU) power off
        Return value: return value from self._PowerControl()
        '''
        return self._PowerControl(POWEROFF)

    def Add(self):
        '''
        Add PCI device(s and their children device if exist)
        Return value(int): success -> 0, error -> 1
        '''
        LogInfo("860 : Add PCIe%s" % self.slot_num)

        # Check slot before power controlling
        status = self._Status()
        if status == POWEROFF:
            pass
        elif status == POWERON:
            LogErr("281 : PCIe%s: device is already available" %
                   self.slot_num)
            return 1
        elif status == EMPTY:
            LogErr("283 : PCIe%s: device is not available" % self.slot_num)
            return 1
        else:
            LogDebug("PCIe%s: error at before power controlling: %s" %
                     (self.slot_num, status))
            LogErr('478 : Internal Error')
            return 1

        # Power controlling
        retval = self.PowerOn()
        if retval == 0:
            pass
        else:
            LogErr("460 : Failed to add PCIe%s with error: %s" %
                   (self.slot_num, retval))
            return 1

        # Confirm the slot is really powered on
        status = self._Status()
        if status == POWERON:
            LogInfo("861 : Added PCIe%s" % self.slot_num)
            return 0
        else:
            if status == POWEROFF or status == EMPTY:
                if status == POWEROFF:
                    slot_status = 'offline'
                else:
                    slot_status = 'empty'
                LogErr("462 : PCIe%s status is invalid: %s" %
                       (self.slot_num, slot_status))
            else:
                LogDebug("PCIe%s: error at after power controlling: %s" %
                         (self.slot_num, status))
                LogErr('479 : Internal Error')
            return 1

    def Remove(self):
        '''
        Remove PCI device(s and their children device if exist)
        Return value(int): success -> 0, error -> 1
        '''
        LogInfo("870 : Remove PCIe%s" % self.slot_num)

        # Check slot before power controlling
        status = self._Status()
        if status == POWEROFF:
            LogErr("288 : PCIe%s: device is already disabled" % self.slot_num)
            return 1
        elif status == POWERON:
            pass
        elif status == EMPTY:
            LogErr("284 : PCIe%s: device is not available" % self.slot_num)
            return 1
        else:
            LogDebug("PCIe%s: error at before power controlling: %s" %
                     (self.slot_num, status))
            LogErr('480 : Internal Error')
            return 1

        # Power controlling
        retval = self.PowerOff()
        if retval == 0:
            pass
        else:
            LogErr("461 : Failed to remove PCIe%s with error %s" %
                   (self.slot_num, retval), 1)
            return 1

        # Confirm the slot is really powered off
        status = self._Status()
        if status == POWEROFF:
            LogInfo("871 : Removed PCIe%s" % self.slot_num)
            return 0
        else:
            if status == POWERON or status == EMPTY:
                if status == POWERON:
                    slot_status = 'online'
                else:
                    slot_status = 'empty'
                LogErr("463 : PCIe%s status is invalid: %s" %
                       (self.slot_num, slot_status))
            else:
                LogDebug("PCIe%s: error at after power controlling: %s" %
                         (self.slot_num, status))
                LogErr('481 : Internal Error')
            return 1


# SB#0 IOU#{0,1,2,3} : 0x100{1,2,3,4}
# SB#1 IOU#{0,1,2,3} : 0x101{1,2,3,4}
# SB#2 IOU#{0,1,2,3} : 0x102{1,2,3,4}
# SB#3 IOU#{0,1,2,3} : 0x103{1,2,3,4}
class IOU(PciHotPlugSlot):
    def __init__(self, dev):
        '''
        And object to manage IOU
        .dev(int): IOU number
        .pci_devs(str): list of all PCI devices (not PCIe only)
        .slot_num(str): PCIe hotplug slot number
        .slot_addr(str): PCI address(seg:bus:dev)
        @dev(int): IOU number
        '''
        self.dev = dev
        self.pci_devs = []
        self.slot_num = ''
        self.slot_addr = ''

    def GetIOUSlotNum(self, dev):
        '''
        Get slot number of IOU and store it as self.slot_num
        @dev(int): IOU number
        Return value(str): SlotNumber of IOU(_SUN) or None
        '''
        if not isinstance(dev, int):
            if isinstance(dev, tuple):
                dev = (dev,)
            LogDebug("%s: wrong dev was passed" % dev)
            return None

        dev = dev + 1  # Last digit of IOU's SlotNumber is +=1 of IOU#
        slot_path = "%s/bus/pci/slots" % DP_SYSFS
        if not os.path.isdir(slot_path):
            LogDebug("%s: not a directory" % slot_path)
            return None
        for entry in os.listdir(slot_path):
            # PCI hotplug driver sets slot name like 'N-M' if slot N
            # has already exists.
            try:
                slot_num = int(entry)
            except ValueError as e:
                LogDebug(e)
                slot_num = int(re.sub(r'(\d+)-\d+', r'\1', entry))
            if slot_num & 0x1000:  # IOU's SlotNumber is 0x10[0-3][1-4]
                if (slot_num & 0x000f) == dev:
                    return entry

    def GetDevsOnIOU(self):
        '''
        Find out all PCI devices under the IOU and store them as self.pci_devs
        Return value: None
        '''
        if not self.slot_num:
            self.slot_num = self.GetIOUSlotNum(self.dev)
        self.slot_addr = self._GetSlotAddr()

        if not self.slot_num or not self.slot_addr:
            return

        pci_devs = []
        pcidev_list = self._ScanPciDevsOnSlot()
        for pci_dev in pcidev_list:
            pci_devs += self.ScanPciDevs(pci_dev)
        self.pci_devs = pci_devs

    def _SetupArg(self, timing, result=''):
        '''
        Set up arguments which are passed to collaboration programs
        @timing(str): Which event this method is called with.
        @result(str): Hotplug result, SUCCESS or FAILURE.  This is set
        when timing is *_POST.
        Return value: list of arguments(str)
        '''
        if not isinstance(timing, str) or not timing in DP_COLLAB_TIMINGS:
            if isinstance(timing, tuple):
                timing = (timing,)
            LogDebug("%s: wrong timing was passed" % timing)
            return []
        if 'POST' in timing:
            if not result or not isinstance(result, str):
                if isinstance(result, tuple):
                    result = (result,)
                LogDebug("%s: wrong result was passed" % result)
                return []
            elif not result == 'SUCCESS' and not result == 'FAILURE':
                LogDebug("%s: wrong result was passed" % result)
                return []

        args = ['-p', "IOU%d" % self.dev]
        args += ['-e', "%s" % timing]

        if timing == 'ADD_POST' or timing == 'RM_POST':
            args += ['-r', "%s" % result]
        elif timing == 'RM_PRE' and self.pci_devs:
            args += ['-d', "%s" % (',').join(self.pci_devs)]

        LogDebug("args: %s" % (' ').join(args))
        return args

    def Show(self, show_slot_num=0):
        '''
        Show PCI/PCIe device(s) info below this IOU
        @show_slot_num(int): if set, show slot number at line head
        Return value(int): 0
        '''
        self.GetDevsOnIOU()
        if not self.pci_devs:
            return 0

        self.pci_devs.sort()
        for pci_dev in self.pci_devs:
            output = CallLspciS(pci_dev)
            sys.stdout.write("%s\n" % output)
        return 0

    def Add(self):
        '''
        Add IOU and its subsidiary PCI devices
        Return value(int):
          success -> 0x0
          error -> 0x1 : invalid device is set
                   0x2 : pre collab call
                   0x5 : iou addition
                   0x7 : post collab call
                   0xc : iou addition and post collab call
        '''
        LogInfo("840 : Add IOU%d" % self.dev)

        # Check slot before power controlling
        self.slot_num = self.GetIOUSlotNum(self.dev)
        status = self._Status()
        if status == POWEROFF:
            pass
        elif status == POWERON:
            LogErr("282 : IOU%d: device is already available" % self.dev)
            return 0x1
        elif status == EMPTY:
            LogErr("285 : IOU%d: device is not available" % self.dev)
            return 0x1
        else:
            LogDebug("IOU%d: error at before power controlling: %s" %
                     (self.dev, status))
            LogErr('482 : Internal Error')
            return 0x1

        collab_ret = 0
        collab_list = _FindCollabs()
        if collab_list:
            collab_arg = self._SetupArg(timing='ADD_PRE')
            collab_ret = CallCollabs("iou%d" % self.dev, 'ADD_PRE', collab_arg,
                                     collab_list)
            if collab_ret:
                return 0x2

        # Power controlling
        retval = self.PowerOn()
        if not retval == 0:
            result = 'FAILURE'
            LogErr("440 : Failed to add IOU%d with error %d" %
                   (self.dev, retval))
        else:
            # Confirm the IOU is really powered on
            status = self._Status()
            if status == POWERON:
                result = 'SUCCESS'
                LogInfo("844: Added IOU%d" % self.dev)
            else:
                if status == POWEROFF or status == EMPTY:
                    if status == POWERON:
                        slot_status = 'offline'
                    else:
                        slot_status = 'empty'
                    LogErr("453: IOU%s status is invalid: %s" %
                           (self.dev, slot_status))
                else:
                    LogDebug("IOU%s: error at after power controlling: %s" %
                             (self.dev, status))
                    LogErr('455 : Internal Error')

        if collab_list:
            collab_arg = self._SetupArg(timing='ADD_POST', result=result)
            collab_ret = CallCollabs("iou%d" % self.dev, 'ADD_POST',
                                     collab_arg, collab_list)
            if collab_ret:
                if result == 'SUCCESS':
                    return 0x7
                else:
                    return 0xc

        if result == 'SUCCESS':
            return 0x0
        else:
            return 0x5

    def Remove(self):
        '''
        Remove IOU and its subsidiary PCI devices
        Return value(int):
          success -> 0x0
          error -> 0x1 : invalid device is set
                   0x2 : pre collab call
                   0x6 : iou removal
                   0x7 : post collab call
                   0xd : iou removal and post collab call
        '''
        LogInfo("850 : Remove IOU%d" % self.dev)

        # Check slot before power controlling
        self.slot_num = self.GetIOUSlotNum(self.dev)
        status = self._Status()
        if status == POWEROFF:
            LogErr("289 : IOU%d: device is already disabled" % self.dev)
            return 0x1
        elif status == POWERON:
            pass
        elif status == EMPTY:
            LogErr("286 : IOU%d: device is not available" % self.dev)
            return 0x1
        else:
            LogDebug("IOU%d: error at before power controlling: %s" %
                     (self.dev, status))
            LogErr('483 : Internal Error')
            return 0x1
        self.GetDevsOnIOU()

        collab_ret = 0
        collab_list = _FindCollabs()
        if collab_list:
            collab_arg = self._SetupArg(timing='RM_PRE')
            collab_ret = CallCollabs("iou%d" % self.dev, 'RM_PRE', collab_arg,
                                     collab_list)
            if collab_ret:
                return 0x2

        # Power controlling
        retval = self.PowerOff()
        if not retval == 0:
            result = 'FAILURE'
            LogErr("450 : Failed to remove IOU%d with error %d" %
                   (self.dev, retval))
        else:
            # Confirm the IOU is really powered off
            status = self._Status()
            if status == POWEROFF:
                result = 'SUCCESS'
                LogInfo("854 : Removed IOU%d" % self.dev)
            else:
                if status == POWERON or status == EMPTY:
                    if status == POWEROFF:
                        slot_status = 'online'
                    else:
                        slot_status = 'empty'
                    LogErr("454: IOU%s status is invalid: %s" %
                           (self.dev, slot_status))
                else:
                    LogDebug("IOU%s: error at after power controlling: %s" %
                             (self.dev, status))
                    LogErr('456 : Internal Error')

        if collab_list:
            collab_arg = self._SetupArg(timing='RM_POST', result=result)
            collab_ret = CallCollabs("iou%d" % self.dev, 'RM_POST', collab_arg,
                                     collab_list)
            if collab_ret:
                if result == 'SUCCESS':
                    return 0x7
                else:
                    return 0xd

        if result == 'SUCCESS':
            return 0x0
        else:
            return 0x6


def CallLspciS(pci_addr, optarg=''):
    '''
    Call 'lspci [optarg] -s pciaddr' to get PCI device info.
    @pci_addr(str): PCI address for target device in seg:bus:dev.fn format
    @optarg(str): optional argument for lspci command
    Return value(str): output of lspci command or None
    '''
    import shlex

    if not isinstance(pci_addr, str) or not pci_addr_regexp.match(pci_addr):
        if isinstance(pci_addr, tuple):
            pci_addr = (pci_addr,)
        LogDebug("%s: wrong pci_addr was passed" % pci_addr)
        return
    if optarg and not isinstance(optarg, str):
        if isinstance(optarg, tuple):
            optarg = (optarg,)
        LogDebug("%s: wrong optarg was passed" % optarg)
        return

    lspci_path = '/sbin/lspci'
    if IsNonExecutable(lspci_path):
        return

    args = [lspci_path, '-s', pci_addr]
    if optarg:
        opt_list = shlex.split(optarg)
        args += opt_list
    LogDebug("Calling lspci: %s" % (' ').join(args))
    try:
        process = subprocess.Popen(args, stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE)
        process.wait()
    except Exception as e:
        LogDebug(e)
        return
    else:
        out_msgs, err_msgs = process.communicate()
        LogDebug("lspci returns with returncode: %d" % process.returncode)
        if out_msgs:
            for out_msg in out_msgs.splitlines():
                LogDebug("%s" % out_msg)
        if err_msgs:
            for err_msg in err_msgs.splitlines():
                LogDebug("%s" % err_msg)
        if process.returncode == 0:
            return out_msgs.rstrip('\n\r')
        else:
            return


class Collab(object):
    def __init__(self, prog, verbose=0, timeout=5):
        '''
        An object to manage Collaboration program
        .prog(str): Collaboration program name
        .verbose(int): a switch for verbose mode
        .timeout(int): waiting a time(minutes) for self.prog
        @prog(str): Collaboration program name
        @verbose(int): a switch for verbose mode
        @timeout(int): waiting a time(minutes) for self.prog
        '''
        self.prog = prog
        self.verbose = verbose
        self.timeout = timeout

    def _MsgHeader(self):
        '''
        Generate a header(in syslog term, tag.  See logger(1))
        Return value(str): A string with: current time in syslog
        format, hostname, dr command name, collab prog name
        '''
        now = time.strftime("%b %d %H:%M:%S", time.localtime())
        # strftime returns the day with %02d format.  However, syslog
        # outputs the day with %2d format.  We can't change latter, so
        # convert %02d to %2d if the day is between 01 and 09.
        # Sample: (Nov 0)9 12:49:22 -> (Nov  )9 12:49:22
        now = re.sub(r'(^[A-z]{3}\s)0', r'\1 ', now)
        try:
            # get hostname, don't need a domain name
            hostname = socket.gethostname().split('.')[0]
        except Exception as e:
            LogDebug("%s: use 'unknown' instead" % e)
            hostname = 'unknown'
        return "%s %s dr-util : %s" % (now, hostname, self.prog)

    def CallLogging(self, read_fd, write_fd, loglevel='INFO'):
        '''
        Logging a message
        @read_fd(file): fd to read from program outputs
        @write_fd(file): fd to write program outputs
        @loglevel(str): logging level, INFO or ERR
        Return value: None
        '''
        if not read_fd or not isinstance(read_fd, file):
            if isinstance(read_fd, tuple):
                read_fd = (read_fd,)
            LogDebug("%s: wrong read_fd was passed" % read_fd)
            return
        if write_fd and not isinstance(write_fd, file):
            if isinstance(write_fd, tuple):
                write_fd = (write_fd,)
            LogDebug("%s: wrong write_fd was passed" % write_fd)
            return
        if not isinstance(loglevel, str):
            if isinstance(loglevel, tuple):
                loglevel = (loglevel,)
            LogDebug("%s: wrong loglevel was passed" % loglevel)
            return
        if not loglevel == 'INFO' and not loglevel == 'ERR':
            LogDebug("%s: wrong loglevel was passed" % loglevel)
            return

        try:
            lines = read_fd.readlines()
        except IOError as e:
            LogDebug(e)
            return

        for line in lines:
            if line == '':  # EOF
                break
            else:
                if write_fd:
                    try:
                        write_fd.write("%s : %s : %s" %
                                       (self._MsgHeader(), loglevel, line))
                    except IOError as e:
                        LogDebug(e)
                        return
                else:
                    if loglevel == 'INFO':
                        priority = syslog.LOG_INFO
                    else:
                        priority = syslog.LOG_ERR
                    Log("%s : %s : %s" % (self.prog, loglevel, line), priority)

    def Call(self, args):
        '''
        Call collaboration program
        @args(str): argument list for the program
        Return Value(int): return value from collaboration program or None
        '''
        import select

        if not args or not isinstance(args, list):
            return errno.EINVAL

        # subprocess.PIPE does not return EPOLLOUT/EPOLLPRI while
        # subprocess is running, so can not get any outputs in sequence.
        # Therefore, try to read subprocess's outputs via temporary files.
        collab_stdout_path = "%s/collab_stdout.txt" % DP_LOG_PATH
        collab_stderr_path = "%s/collab_stderr.txt" % DP_LOG_PATH
        try:
            collab_out_write = open(collab_stdout_path, 'w')
            collab_out_read = open(collab_stdout_path, 'r')
            collab_err_write = open(collab_stderr_path, 'w')
            collab_err_read = open(collab_stderr_path, 'r')
        except OSError as e:
            LogDebug(e)
            return e.errno
        except Exception as e:
            LogDebug(e)
            return None

        read_poll = select.poll()
        read_poll.register(collab_out_read,
                           select.POLLOUT | select.POLLPRI | select.POLLHUP)
        read_poll.register(collab_err_read,
                           select.POLLOUT | select.POLLPRI | select.POLLHUP)

        fd = None
        if self.verbose:
            args += ['-v']
            log_path = "%s/%s.log" % (DP_LOG_PATH, self.prog)
            try:
                fd = open(log_path, 'a')
            except Exception as e:
                # fallback to default(outputs to syslog)
                LogDebug("%s: fall back to default" % e)

        if self.timeout == 0:
            timeout = 0xffffff
        else:
            timeout = self.timeout * 60
        elapsed_seconds = 0

        command_path = "%s/%s" % (DP_USER_COMMAND_PATH, self.prog)
        retval = IsNonExecutable(command_path)
        if retval:
            if retval == errno.EACCES:
                mode = oct(os.stat(command_path).st_mode & 0777)
                LogErr("622 : %s : Permission denied, %s" %
                       (self.prog, mode))
            if fd:
                fd.close()
            return retval
        command = [command_path] + args
        collab_prog = subprocess.Popen(command, stdout=collab_out_write,
                                       stderr=collab_err_write, close_fds=True)

        while elapsed_seconds < timeout:
            # check collab_prog is still running or not
            returncode = collab_prog.poll()
            if returncode is None:
                # Not finished yet
                pass
            else:
                # collab_prog has finished
                stdout, stderr = read_poll.poll(1)
                if stdout:
                    # collab_prog outputs messages to stdout
                    self.CallLogging(collab_out_read, fd, 'INFO')
                if stderr:
                    # collab_prog outputs messages to stderr
                    self.CallLogging(collab_err_read, fd, 'ERR')

                collab_out_read.close()
                collab_err_read.close()
                if fd:
                    fd.close()
                return returncode

            stdout, stderr = read_poll.poll(1)
            if stdout:
                # collab_prog outputs messages to stdout
                self.CallLogging(collab_out_read, fd, 'INFO')
            if stderr:
                # collab_prog outputs messages to stderr
                self.CallLogging(collab_err_read, fd, 'ERR')

            time.sleep(1)
            if not self.timeout == 0:  # do not timeout if self.timeout == 0
                elapsed_seconds += 1

        else:
            # collab_prog did not finish within self.timeout
            collab_prog.terminate()  # send SIGTERM to the collab_prog
            # reset the timer
            elapsed_seconds = 0
            timeout = 60
            while elapsed_seconds < timeout:
                # check collab_prog is still running or not
                returncode = collab_prog.poll()
                if returncode is None:
                    # Not finished yet
                    pass
                else:
                    # collab_prog has finished
                    stdout, stderr = read_poll.poll(1)
                    if stdout:
                        # collab_prog outputs messages to stdout
                        self.CallLogging(collab_out_read, fd, 'INFO')
                    if stderr:
                        # collab_prog outputs messages to stderr
                        self.CallLogging(collab_err_read, fd, 'ERR')

                    collab_out_read.close()
                    collab_err_read.close()
                    if fd:
                        fd.close()
                    return returncode

                stdout, stderr = read_poll.poll(1)
                if stdout:
                    # collab_prog outputs messages to stdout
                    self.CallLogging(collab_out_read, fd, 'INFO')
                if stderr:
                    # collab_prog outputs messages to stderr
                    self.CallLogging(collab_err_read, fd, 'ERR')

                time.sleep(1)
                elapsed_seconds += 1
            else:
                # collab_prog did not finish within the extra 60 secs
                # Can not wait any more, 'Hasta la vista, baby.'
                collab_prog.kill()
                returncode = collab_prog.poll()
                if returncode is None:
                    # Not finished yet, wait 10 more seconds
                    time.sleep(10)
                stdout, stderr = read_poll.poll(1)
                if stdout:
                    # collab_prog outputs messages to stdout
                    self.CallLogging(collab_out_read, fd, 'INFO')
                if stderr:
                    # collab_prog outputs messages to stderr
                    self.CallLogging(collab_err_read, fd, 'ERR')

                collab_out_read.close()
                collab_err_read.close()
                if fd:
                    fd.close()
                return returncode


def _FindCollabs():
    '''
    Find out installed collaboration program
    Return value: list of class Collab object.
    Note: The list order is incertitude.  Need to sort if you care the
    order like ascending or descending.
    '''
    collab_list = []
    collab_regexp = re.compile(r'\d{2}-(.*)')
    verbose_regexp = re.compile(r'^\s*verbose\s*=\s*([a-z]+)[^a-z]*',
                                re.IGNORECASE)
    timeout_regexp = re.compile(r'^\s*timeout\s*=\s*([\d]+)[^\d]*',
                                re.IGNORECASE)

    if not os.path.isdir(DP_USER_COMMAND_PATH):
        LogDebug("%s: not a directory" % DP_USER_COMMAND_PATH)
        return []
    for entry in os.listdir(DP_USER_COMMAND_PATH):
        if not collab_regexp.match(entry):
            continue
        conf_path = "%s/%s.conf" % (DP_CONF_PATH,
                                    collab_regexp.sub(r'\1', entry))
        if not os.path.exists(conf_path):
            collab = Collab(entry)
        else:
            lines, error = ReadlinesFile(conf_path)
            if not lines:
                collab = Collab(entry)
            else:
                verbose = 'false'
                timeout = 5
                # If multiple statements are in a file, the last one
                # becomes effective
                for line in lines:
                    if verbose_regexp.match(line):
                        verbose = verbose_regexp.sub(r'\1', line).lower()
                    elif timeout_regexp.match(line):
                        timeout = int(timeout_regexp.sub(r'\1', line))

                if timeout < 0 or 0 < timeout < 5 or timeout > 1024:
                    # these are out of range. use default value instead
                    timeout = 5
                if verbose == 'true':
                    collab = Collab(entry, 1, timeout)
                else:
                    collab = Collab(entry, 0, timeout)

        collab_list.append(collab)
    return collab_list


def _GetCallCollabsEventId(device, timing):
    '''
    Get event ID from device name and timing for CallCollabs()
    @device(str): Hot add{ing,ed}/remov{ing,ed} device (sbX or iouX)
    @timing(str): Calling timing {ADD,RM}_{PRE,POST}
    Return value(int): from 0 to 4 (0 means no match)
    '''
    if not isinstance(device, str):
        if isinstance(device, tuple):
            device = (device,)
        LogDebug("%s: wrong device was passed" % device)
        return 0
    if not 'sb' in device and not 'iou' in device:
        LogDebug("%s: wrong device was passed" % device)
        return 0
    if not isinstance(timing, str) or not timing in DP_COLLAB_TIMINGS:
        if isinstance(timing, tuple):
            timing = (timing,)
        LogDebug("%s: wrong timing was passed" % timing)
        return 0

    event_id = 0
    if 'sb' in device:
        if 'ADD' in timing:
            event_id = 1
        elif 'RM' in timing:
            event_id = 2
    elif 'iou' in device:
        if 'ADD' in timing:
            event_id = 3
        elif 'RM' in timing:
            event_id = 4
    return event_id


def CallCollabs(device, timing, args, collab_list):
    '''
    Call collaboration program to notify system resource changing
    @device(str): Hot add{ing,ed}/remov{ing,ed} device (sbX or iouX)
    @timing(str): Calling timing {ADD,RM}_{PRE,POST}
    @args(str): list of command arguments to be passed to
    collaboration programs
    @collab_list: List of Collab object
    Return Value(int): success (include error but sustainable) -> 0,
    error -> 1
    '''
    if not isinstance(device, str):
        if isinstance(device, tuple):
            device = (device,)
        LogDebug("%s: wrong device was passed" % device)
        return 1
    if not 'sb' in device and not 'iou' in device:
        LogDebug("%s: wrong device was passed" % device)
        return 1
    if not isinstance(timing, str) or not timing in DP_COLLAB_TIMINGS:
        if isinstance(timing, tuple):
            timing = (timing,)
        LogDebug("%s: wrong timing was passed" % timing)
        return 1
    if not isinstance(args, list):
        if isinstance(args, tuple):
            args = (args,)
        LogDebug("%s: wrong args was passed" % args)
        return 1
    if not collab_list is None and not isinstance(collab_list, list):
        if isinstance(collab_list, tuple):
            collab_list = (collab_list,)
        LogDebug("%s: wrong collab_list was passed" % collab_list)
        return 1

    # The order of items is Error, SB ADD, SB RM, IOU ADD, IOU RM
    zero_msg_id = ['None', '806', '826', '841', '851']
    pre_msg_id = ['None', '807', '827', '842', '852']
    post_msg_id = ['None', '808', '828', '843', '853']
    fail_msg_id = ['None', '404', '425', '441', '451']
    event_id = _GetCallCollabsEventId(device, timing)

    if not collab_list:
        LogInfo("%s : Execute 0 user programs at %s timing" %
                (zero_msg_id[event_id], timing))
        return 0

    # Sort collab_list in reverse (or normal) alphabetical order
    if 'FAILURE' in args:
        collab_list.sort(lambda x, y: cmp(x.prog, y.prog), reverse=True)
    else:
        collab_list.sort(lambda x, y: cmp(x.prog, y.prog))

    LogInfo("%s : Execute %d user programs at %s timing" %
            (pre_msg_id[event_id], len(collab_list), timing))

    for collab in collab_list:
        LogDebug("Calling: %s, v=%s, t=%s" %
                 (collab.prog, collab.verbose, collab.timeout))
        retval = collab.Call(args)
        LogDebug("Finished: %s, v=%s, t=%s" %
                 (collab.prog, collab.verbose, collab.timeout))
        if not retval == 0:  # some error happened
            LogErr("%s : Exited %s with status %s at %s timing" %
                   (fail_msg_id[event_id], collab.prog, str(retval), timing))
            return 1

    LogInfo("%s : Executed user programs at %s timing" %
            (post_msg_id[event_id], timing))
    return 0


def PrintSBsStatus():
    '''
    Print all SB status
     Format: SBx: {online, online* or empty}
    Return value: success -> 0, error -> 1
    '''
    sb_status = ['empty'] * DP_SB_MAX

    retval = 0
    path = "%s/bus/acpi/devices" % DP_SYSFS
    if not os.path.isdir(path):
        LogDebug("%s: not a directory" % path)
        retval = 1
    else:
        for entry in os.listdir(path):
            if not 'ACPI0004:' in entry:
                continue
            # On RHEL7.1 or later, ACPI0004 directory always exists.
            # To check ACPI0004 is poweron or not, the status file must
            # be checked on RHEL7.1 or later
            sta_path = "%s/%s/status" % (path, entry)
            if os.path.exists(sta_path):
                sta, error = ReadFile(sta_path)
                if sta is None:
                    retval = 1
                    continue
                if int(sta) == 0:
                    continue
            sun_path = "%s/%s/sun" % (path, entry)
            sun, error = ReadFile(sun_path)
            if sun is None:
                retval = 1
                continue
            sb_num = int(sun)
            sb = SB(sb_num)
            sb._FindNodes()
            if sb.HasNonMovableZone():
                status = 'online'
            else:
                status = 'online*'
            sb_status[sb_num] = status

    for sb_num in range(0, DP_SB_MAX):
        sys.stdout.write("SB%d: %s\n" % (sb_num, sb_status[sb_num]))
    return retval


def PrintCPUsStatus():
    '''
    Print all CPU status
     Format: cpux: {online, offline, unknown}
    Return value: success -> 0, error -> 1
    '''
    cpu_list = []  # list of CPU number in string format
    cpu_regexp = re.compile(r'^cpu([\d]+)$')

    cpu_path = "%s/devices/system/cpu" % DP_SYSFS
    if not os.path.isdir(cpu_path):
        LogDebug("%s: not a directory" % cpu_path)
        return 1
    for entry in os.listdir(cpu_path):
        if not cpu_regexp.match(entry):
            continue
        cpu_num = cpu_regexp.sub(r'\1', entry)
        cpu_list.append(cpu_num)
    else:
        cpu_list.sort(lambda x, y: cmp(int(x), int(y)))

    retval = 0
    for cpu in cpu_list:
        power_path = "%s/cpu%s/online" % (cpu_path, cpu)
        power, error = ReadFile(power_path)
        if power is None:
            if error == errno.ENOENT and cpu == '0':
                # on this system, BSP does not have hotplug capability
                # and is always online
                sys.stdout.write('cpu0: online\n')
            else:
                sys.stdout.write("cpu%s: unknown\n" % cpu)
                retval = 1
        elif power == str(POWERON):
            sys.stdout.write("cpu%s: online\n" % cpu)
        else:
            sys.stdout.write("cpu%s: offline\n" % cpu)
    return retval


def PrintMEMsStatus():
    '''
    Print all MEM status
     Format: memoryx: {online, offline, unknown}
    Return value: success -> 0, error -> 1
    '''
    mem_list = []  # list of MEM number in string format
    mem_regexp = re.compile(r'^memory([\d]+)$')

    path = "%s/devices/system/memory" % DP_SYSFS
    if not os.path.isdir(path):
        LogDebug("%s: not a directory" % path)
        return 1
    for entry in os.listdir(path):
        if not mem_regexp.match(entry):
            continue
        mem_num = mem_regexp.sub(r'\1', entry)
        mem_list.append(mem_num)
    else:
        mem_list.sort(lambda x, y: cmp(int(x), int(y)))

    retval = 0
    for mem in mem_list:
        power_path = "%s/memory%s/state" % (path, mem)
        power, error = ReadFile(power_path)
        if power is None:
            sys.stdout.write("memory%s: unknown\n" % mem)
            retval = 1
        else:
            sys.stdout.write("memory%s: %s\n" % (mem, power))
    return retval


def _PrintIODevsStatus(dev_type):
    '''
    Print all IOU/PCIe status
    @dev_type(str): IOU or pcie
    Return value: success -> 0, error -> 1

    Status:
      online:  (adapter, power) = (1, 1)
      offline: (adapter, power) = (1, 0)
      empty:   (adapter, power) = (0, 0)
    '''
    if not dev_type == 'IOU' and not dev_type == 'pcie':
        if isinstance(dev_type, tuple):
            dev_type = (dev_type,)
        LogDebug("%s: wrong dev_type was passed" % dev_type)
        return 1

    dev_list = []  # list of PCIe slot number in string format

    path = "%s/bus/pci/slots" % DP_SYSFS
    if not os.path.isdir(path):
        LogDebug("%s: not a directory" % path)
        return 1
    for entry in os.listdir(path):
        # PCI hotplug driver sets slot name like 'N-M' if slot N has
        # already exists.
        try:
            slot_num = int(entry)
        except ValueError as e:
            LogDebug(e)
            slot_num = int(re.sub(r'(\d+)-\d+', r'\1', entry))
        if dev_type == 'IOU':
            if not slot_num & 0x1000:  # IOU's SlotNumber is 0x10[0-3][1-4]
                continue
        else:
            if slot_num & 0x1000:
                continue
        dev_list.append(entry)
    else:
        dev_list = ListSort(dev_list)

    retval = 0
    for dev in dev_list:
        try:
            dev_num = int(dev)
        except ValueError as e:
            LogDebug(e)
            dev_num = int(re.sub(r'(\d+)-\d+', r'\1', dev))
        if dev_type == 'IOU':
            dev_num = (dev_num & 0x000f) - 1
        power_path = "%s/%s/power" % (path, dev)
        power, error = ReadFile(power_path)
        if power == str(POWERON):
            sys.stdout.write("%s%s: online\n" % (dev_type, dev_num))
            continue

        adapter_path = "%s/%s/adapter" % (path, dev)
        adapter, error = ReadFile(adapter_path)
        if adapter is None:
            sys.stdout.write("%s%s: unknown\n" % (dev_type, dev_num))
            retval = 1
        elif adapter == str(POWERON):
            sys.stdout.write("%s%s: offline\n" % (dev_type, dev_num))
        else:
            sys.stdout.write("%s%s: empty\n" % (dev_type, dev_num))
    return retval


def PrintIOUsStatus():
    '''
    Wrapper method for showing IOUs Status
    Return value: return value from _PrintIODevsStatus()
    '''
    return _PrintIODevsStatus('IOU')


def PrintPCIesStatus():
    '''
    Wrapper method for showing PCIes Status
    Return value: return value from _PrintIODevsStatus()
    '''
    return _PrintIODevsStatus('pcie')


def CollectCmdResult(command_line, archive_path):
    '''
    Execute a command and store the command output into archive_path
    @command_line(str): command line
    @archive_path(str): directory path
    Return value(int): success -> 0, error -> 1
    '''
    if not command_line or not isinstance(command_line, str):
        if isinstance(command_line, tuple):
            command_line = (command_line,)
        LogDebug("%s: wrong command_line was passed" % command_line)
        return 1
    if not archive_path or not isinstance(archive_path, str):
        if isinstance(archive_path, tuple):
            archive_path = (archive_path,)
        LogDebug("%s: wrong archive_path was passed" % archive_path)
        return 1
    if not os.path.isdir(archive_path):
        LogDebug("%s: not a directory" % archive_path)
        return 1

    lists = command_line.split()
    command = lists[0]
    if IsNonExecutable(command):
        return 1
    command_base = os.path.basename(command)
    if len(lists) > 1:
        command_args = lists[1:]  # get rid of lists[0] from lists
        output_path = ("%s/%s-%s.txt" %
                       (archive_path, command_base,
                        ('-').join(command_args).replace('/', '-')))
        args = "%s %s > %s 2>&1" % (command, (' ').join(command_args),
                                    output_path)
    else:
        output_path = "%s/%s.txt" % (archive_path, command_base)
        args = "%s > %s 2>&1" % (command, output_path)

    try:
        process = subprocess.Popen(args, shell=True, stderr=subprocess.PIPE)
        process.wait()
    except Exception as e:
        LogDebug(e)
        return 1
    else:
        err_msgs = process.communicate()[1]
        LogDebug("%s returns with returncode: %d" % (args, process.returncode))
        if err_msgs:
            for err_msg in err_msgs.splitlines():
                sys.stderr.write("%s\n" % err_msg)
                LogDebug("%s" % err_msg)
        if process.returncode == 0:
            return 0
        else:
            return 1


def CollectFiles(file_path, archive_path):
    '''
    Copy a file/directory to archive_path
    @file_path(str): path to a file or directory
    @archive_path(str): directory path
    Return value(int): success -> 0, error -> 1
    '''
    import shutil

    if not file_path or not isinstance(file_path, str):
        if isinstance(file_path, tuple):
            file_path = (file_path,)
        LogDebug("%s: wrong file_path was passed" % file_path)
        return 1
    if not archive_path or not isinstance(archive_path, str):
        if isinstance(archive_path, tuple):
            archive_path = (archive_path,)
        LogDebug("%s: wrong archive_path was passed" % archive_path)
        return 1
    if not os.path.isdir(archive_path):
        LogDebug("%s: not a directory" % archive_path)
        return 1

    archive_path = archive_path.rstrip('/')  # removing trailing '/' if exists

    retval = 1
    if os.path.isfile(file_path):  # copiable file
        try:
            shutil.copy2(file_path, archive_path)
        except Exception as e:
            LogDebug(e)
        else:
            retval = 0
    elif os.path.isdir(file_path):
        file_path = file_path.rstrip('/')
        copied_dir = os.path.basename(file_path)
        try:
            shutil.copytree(file_path, "%s/%s" % (archive_path, copied_dir),
                            symlinks=True)
        except shutil.Error as e:
            for error in e.args[0]:
                LogDebug(error[2])  # logging exception only
        except Exception as e:
            LogDebug(e)
        else:
            retval = 0
    else:  # not copiable file or not exist
        retval = 1
    return retval


def Report():
    '''
    Collect up information related to DR command and archive it.
    Return value(int): success -> 0, error -> 1
    '''
    from tempfile import mkdtemp
    import shutil

    default = ['C: /bin/dmesg',
               "D: %s/devices/system/node/node.*" % DP_SYSFS,
               'D: /opt/FJSVdr-util/var/log',
               'D: /opt/FJSVdr-util/etc',
               'D: /opt/FJSVdr-util/user_command',
               "D: %s/zoneinfo" % DP_PROCFS,
               'C: /sbin/lspci',
               'C: /sbin/lspci -vvv']
    appendix = []
    command_regexp = re.compile(r'^([CDF]):\s*(.*)\s*$')
    comment_regexp = re.compile(r'^\s*#')
    blank_regexp = re.compile(r'^\s*$')

    try:
        # get hostname, don't need a domain name
        hostname = socket.gethostname().split('.')[0]
    except Exception as e:
        LogDebug("%s: use 'unknown' instead" % e)
        hostname = 'unknown'
    now = time.strftime("%Y%m%d%H%M%S", time.localtime())  # 20121109-172600
    archive_prefix = "drreport-%s-%s" % (hostname, now)
    try:
        archive_path = mkdtemp(prefix=archive_prefix)
    except OSError as e:
        if e.errno == errno.ENOSPC:
            LogErr('101 : No space left on device')
        else:
            LogDebug(e)
            LogErr("102 : %s" % e)
        return 1
    except Exception as e:
        LogDebug(e)
        LogErr('300: Internal Error')
        return 1
    archive_dir = archive_path.split('/')[-1]

    report_conf_path = "%s/dr_report.conf" % DP_CONF_PATH
    lines, error = ReadlinesFile(report_conf_path)
    for line in lines:
        if blank_regexp.match(line) or comment_regexp.match(line):
            continue
        if not command_regexp.match(line):
            LogDebug("Invalid format, ignore this line: %s" % line)
            continue
        appendix.append(line)

    commands = default + appendix
    for command in commands:
        command_type = command_regexp.sub(r'\1', command)
        command_body = command_regexp.sub(r'\2', command)

        if command_type == 'C':
            # execute command
            CollectCmdResult(command_body, archive_path)
        elif command_type == 'D' or command_type == 'F':
            # copy files
            CollectFiles(command_body, archive_path)
        else:
            # ignore the command
            pass

    tar_path = '/bin/tar'
    if IsNonExecutable(tar_path):
        return 1
    output_file = "%s.tar.xz" % archive_path
    args = [tar_path, 'Jc', '-C', "%s/.." % archive_path, '-f', output_file,
            archive_dir]
    try:
        process = subprocess.Popen(args, stderr=subprocess.PIPE)
        process.wait()
    except Exception as e:
        LogDebug("%s: %s" % (e, args))
        LogErr('301 : Internal Error')
        return 1
    else:
        err_msgs = process.communicate()[1]
        LogDebug("%s returns with returncode: %d" % (args, process.returncode))
        if err_msgs:
            for err_msg in err_msgs.splitlines():
                LogErr("103 : %s" % err_msg)
        if process.returncode == 0:
            LogInfo("880 : Create report file at %s" % output_file, console=1)
            try:
                shutil.rmtree(archive_path)
            except Exception as e:
                LogDebug(e)
            return 0
        else:
            return 1


def Log(msg, priority=syslog.LOG_INFO, console=0):
    '''
    Output a message to syslog and console(if needed)
    @msg(str): a message to be shown
    @priority(int): log level
    @console(int): 0 -> do not show on console, 1 -> show on console
    Return value: None
    '''
    if not isinstance(priority, int):
        if isinstance(priority, tuple):
            priority = (priority,)
        LogDebug("%s: wrong priority was passed, but falls back as "
                 'syslog.LOG_INFO' % priority)
        priority = syslog.LOG_INFO
    if not syslog.LOG_EMERG <= priority <= syslog.LOG_DEBUG:
        LogDebug("%s: wrong priority was passed, but falls back as "
                 'syslog.LOG_INFO' % priority)
        priority = syslog.LOG_INFO
    if not isinstance(console, int) or not 0 <= console <= 1:
        if isinstance(console, tuple):
            console = (console,)
        LogDebug("%s: wrong console was passed, but falls back as console=0"
                 % console)
        console = 0

    msg = str(msg)  # convert all inputs to String
    logopt = syslog.LOG_PID
    syslog.openlog('dr-util', logopt)
    syslog.syslog(priority, msg)

    if console:
        pid = os.getpid()
        msg_header = "dr-util[%d]: " % pid
        msg = "%s%s\n" % (msg_header, msg)
        if priority == syslog.LOG_INFO or priority == syslog.LOG_DEBUG:
            sys.stdout.write(msg)
        else:
            sys.stderr.write(msg)


def LogEmerg(msg, console=1):
    Log("EMERG : %s" % msg, syslog.LOG_EMERG, console)


def LogAlert(msg, console=1):
    Log("ALERT : %s" % msg, syslog.LOG_ALERT, console)


def LogCrit(msg, console=1):
    Log("CRIT : %s" % msg, syslog.LOG_CRIT, console)


def LogErr(msg, console=1):
    Log("ERR : %s" % msg, syslog.LOG_ERR, console)


def LogWarning(msg, console=1):
    Log("WARNING : %s" % msg, syslog.LOG_WARNING, console)


def LogNotice(msg, console=1):
    Log("NOTICE : %s" % msg, syslog.LOG_NOTICE, console)


def LogInfo(msg, console=0):
    Log("INFO : %s" % msg, syslog.LOG_INFO, console)


def LogDebug(msg, console=0):
    if DP_DEBUG:
        Log("DEBUG : %s" % msg, syslog.LOG_DEBUG, console)

    path = "%s/dr_debug.log" % DP_LOG_PATH
    try:
        fd = open(path, 'a')
    except Exception:
        return
    else:
        try:
            with fd:
                frame = None
                filename = None
                lineno = None
                try:
                    now = time.strftime("%b %d %H:%M:%S", time.localtime())
                    # get caller's frame and debug info
                    frame = inspect.getframeinfo(inspect.currentframe().f_back)
                    filename = frame.filename
                    lineno = frame.lineno
                except Exception:
                    pass
                fd.write("%s: %s(%s): %s\n" % (now, filename, lineno, msg))
        except Exception:
            pass
        return


def AvoidOom():
    '''
    Not to be killed by OOM, reset oom_score_adj
    Return value(int): success -> 0, error -> 1
    '''
    path = "%s/self/oom_score_adj" % DP_PROCFS
    retval = WriteFile(path, '0')
    if retval:
        return 1
    else:
        return 0


def ReadFile(path, mode='r'):
    '''
    Read whole contents from a file.
    @path(str): an absolute file path to be read
    @mode(str): file open mode (must be .*r.*)
    Return value(tuple (str, int)): contents of file or None if error
    happens and success -> 0, error -> errno
    '''
    if not path or not isinstance(path, str):
        if isinstance(path, tuple):
            path = (path,)
        LogDebug("%s: wrong path was passed" % path)
        return (None, errno.EINVAL)
    if not mode or not isinstance(mode, str) or not 'r' in mode:
        if isinstance(mode, tuple):
            mode = (mode,)
        LogDebug("%s: wrong mode was passed" % mode)
        return (None, errno.EINVAL)

    try:
        fd = open(path, mode)
    except Exception as e:
        LogDebug(e)
        return (None, e.errno)
    else:
        try:
            with fd:
                contents = None
                contents = fd.read()
        except Exception as e:
            LogDebug(e)
            return (None, e.errno)
        return (contents.rstrip('\n\r'), 0)


def ReadlinesFile(path, mode='r'):
    '''
    Read whole contents from a file.
    @path(str): an absolute file path to be read
    @mode(str): file open mode (must be .*r.*)
    Return value(tuple (list, int)): a list of lines of file or None
    if error happens and success -> 0, error -> errno
    '''
    if not path or not isinstance(path, str):
        if isinstance(path, tuple):
            path = (path,)
        LogDebug("%s: wrong path was passed" % path)
        return (None, errno.EINVAL)
    if not mode or not isinstance(mode, str) or not 'r' in mode:
        if isinstance(mode, tuple):
            mode = (mode,)
        LogDebug("%s: wrong mode was passed" % mode)
        return (None, errno.EINVAL)

    try:
        fd = open(path, mode)
    except Exception as e:
        LogDebug(e)
        return (None, e.errno)
    else:
        try:
            with fd:
                contents = []
                contents = fd.readlines()
        except Exception as e:
            LogDebug(e)
            return (None, e.errno)
        return (contents, 0)


def WriteFile(path, value, mode='w'):
    '''
    Write value to a file.
    @path(str): an absolute file path to be written
    @value(str): value to be written
    @mode(str): file open mode (must be .*w.* or .*a.*)
    Return value(int): success -> 0, error -> errno
    '''
    mode_regexp = re.compile(r'[wa]')

    if not path or not isinstance(path, str):
        if isinstance(path, tuple):
            path = (path,)
        LogDebug("%s: wrong path was passed" % path)
        return errno.EINVAL
    if not isinstance(mode, str) or not mode_regexp.search(mode):
        if isinstance(mode, tuple):
            mode = (mode,)
        LogDebug("%s: wrong mode was passed" % mode)
        return errno.EINVAL

    regexp = re.compile(r'.*\n')
    try:
        fd = open(path, mode)
    except Exception as e:
        LogDebug(e)
        return e.errno
    else:
        try:
            with fd:
                value = str(value)  # convert all inputs to String
                if not regexp.match(value):
                    value = "%s\n" % value
                fd.write(value)
        except Exception as e:
            LogDebug(e)
            return e.errno
        else:
            return 0


def ScanPciDevsOnSystem():
    '''
    Get list of PCI devices on this system.
    Return value(list of str): sorted list of PCI addresses (seg:bus:dev.fn)
    '''
    path = "%s/bus/pci/devices" % DP_SYSFS
    if not os.path.isdir(path):
        LogDebug("%s: not a directory" % path)
        return []

    pci_devs = []
    for entry in os.listdir(path):
        if pci_addr_regexp.match(entry):
            pci_devs.append(entry)

    pci_devs.sort()
    return pci_devs


def IsNonExecutable(cmd_path):
    '''
    Check a command is executable(exists and has os.X_OK.) or not.
    @cmd_path(str): command path
    Return value(int): Executable -> 0, not executable ->
    errno.{EINVAL,ENOENT,EACCES}
    '''
    if not cmd_path or not isinstance(cmd_path, str):
        if isinstance(cmd_path, tuple):
            cmd_path = (cmd_path,)
        LogDebug("%s: wrong cmd_path was passed" % cmd_path)
        return errno.EINVAL

    if not os.path.exists(cmd_path):
        LogDebug("%s: No such file or directory" % cmd_path)
        return errno.ENOENT
    if not os.access(cmd_path, os.X_OK):
        LogDebug("%s: Permission denied" % cmd_path)
        return errno.EACCES
    return 0


def ListSort(num_list):
    '''
    Sort list which may include numbers like M-N format.
    @num_list(str): list of numbers like ['0', '5', '2', '1-1', '1'].
    Return value(list of str): sorted list like ['0', '1', '1-1', '2',
    '5'] or []
    '''
    hyphen_regexp = re.compile(r'(\d+)-(\d+)')

    if not num_list or not isinstance(num_list, list):
        if isinstance(num_list, tuple):
            num_list = (num_list,)
        LogDebug("%s: wrong num_list was passed" % num_list)
        return []
    for item in num_list:
        try:
            int(item)
        except ValueError:
            if hyphen_regexp.match(item):
                continue
            LogDebug("%s: unpexptected item in the num_list" % item)
            return []

    if len(num_list) == 1:
        return num_list

    for head in range(0, len(num_list)):
        rest_items = range(head, len(num_list))
        rest_items.reverse()
        for foot in rest_items:
            prev_prefix = int(hyphen_regexp.sub(r'\1', num_list[foot - 1]))
            next_prefix = int(hyphen_regexp.sub(r'\1', num_list[foot]))
            if prev_prefix == next_prefix:
                if '-' in num_list[foot - 1] and '-' not in num_list[foot]:
                    tmp = num_list[foot]
                    num_list[foot] = num_list[foot - 1]
                    num_list[foot - 1] = tmp
                elif '-' in num_list[foot - 1] and '-' in num_list[foot]:
                    prev_suffix = int(hyphen_regexp.sub(r'\2',
                                                        num_list[foot - 1]))
                    next_suffix = int(hyphen_regexp.sub(r'\2', num_list[foot]))
                    if prev_suffix > next_suffix:
                        tmp = num_list[foot]
                        num_list[foot] = num_list[foot - 1]
                        num_list[foot - 1] = tmp
            elif prev_prefix > next_prefix:
                tmp = num_list[foot]
                num_list[foot] = num_list[foot - 1]
                num_list[foot - 1] = tmp
    return num_list


def ListToAbbrev(num_list):
    '''
    Convert list of numbers to the abbrev. format
    @numlist(str): list of numbers like ['0', '1', '2', '3', '4'].
    Return Value(tuple (str, int)): Tuple of 'abbreviated string' or
    None, and number of numbers in strlist.  ('0-4', 5)
    Note: No need to sort() before call this method.
    '''
    if not num_list or not isinstance(num_list, list) or '' in num_list:
        if isinstance(num_list, tuple):
            num_list = (num_list,)
        LogDebug("%s: wrong num_list was passed" % num_list)
        return (None, 0)

    # convert string list to int list for easy comparison
    lists = []
    for item in num_list:
        try:
            lists.append(int(item))
        except ValueError:
            LogDebug("%s: wrong num_list was passed" % num_list)
            return (None, 0)
    lists.sort()

    length = len(lists)
    index = 0
    abbrev = ''
    while index < length:
        try:
            head = lists[index]
            while lists[index + 1] == lists[index] + 1 and index < length:
                index += 1
        except IndexError:  # reached end of list
            if head == lists[index]:
                abbrev += "%d" % head
            elif head == lists[index] - 1:  # serial number
                abbrev += "%d,%d" % (head, lists[index])
            else:
                abbrev += "%d-%d" % (head, lists[index])
            break
        else:
            if head == lists[index]:
                abbrev += "%d" % head
            elif head == lists[index] - 1:  # serial number
                abbrev += "%d,%d" % (head, lists[index])
            else:
                abbrev += "%d-%d" % (head, lists[index])

            index += 1
            abbrev += ','

    return (abbrev, index + 1)


def AbbrevToList(abbrev):
    '''
    Convert abbrev. format to the list of numbers
    @abbrev: list of numbers in abbrev. format like '0-4' (str)
    Return Value(str): list of numbers ['0', '1', '2', '3', '4']
    '''
    abbrev_regexp = re.compile(r'[^\-,\d]')
    abbrev_hyphen_regexp = re.compile(r'\d-\d')

    if not isinstance(abbrev, str) or abbrev_regexp.search(abbrev):
        if isinstance(abbrev, tuple):
            abbrev = (abbrev,)
        LogDebug("%s: wrong abbrev was passed" % abbrev)
        return []

    lists = abbrev.split(',')
    if '' in lists:  # get rid of ',,' case
        return []
    nums = []
    for num_lists in lists:
        if not '-' in num_lists:
            nums.append(num_lists)
        else:
            if not abbrev_hyphen_regexp.search(num_lists):
                # get rid of '-\d....' or '....\d-' case
                return []
            region = num_lists.split('-')
            start = int(region[0])
            end = int(region[1])
            for num in range(start, end + 1):
                nums.append(str(num))
    else:
        return nums


def AbbrevToNum(abbrev):
    '''
    Convert abbrev. format to the number of items
    @abbrev(str): list of numbers in abbrev. format like '0-4,6,8-10'
    Return Value(int): number of items. ex. '0-4' -> 5
    '''
    abbrev_regexp = re.compile(r'[^\-,\d]')
    abbrev_hyphen_regexp = re.compile(r'\d-\d')

    if not isinstance(abbrev, str) or abbrev_regexp.search(abbrev):
        if isinstance(abbrev, tuple):
            abbrev = (abbrev,)
        LogDebug("%s: wrong abbrev was passed" % abbrev)
        return 0

    lists = abbrev.split(',')
    if '' in lists:  # get rid of ',,' case
        return 0
    items = 0
    for num_lists in lists:
        if not '-' in num_lists:
            items += 1
        else:
            if not abbrev_hyphen_regexp.search(num_lists):
                # get rid of '-\d....' or '....\d-' case
                return 0
            nums = num_lists.split('-')
            items += int(nums[1]) - int(nums[0]) + 1
    else:
        return items


def CheckConcurrentRunning():
    '''
    Check other dr command is running
    Return Value(int): not running -> 0, running or error -> 1
    '''
    try:
        fd = os.open(DP_PID_FILE, os.O_WRONLY | os.O_CREAT | os.O_EXCL, 0644)
    except OSError as e:
        if e.errno == errno.EEXIST:
            LogDebug(e)
            another_pid, error = ReadFile(DP_PID_FILE)
            LogErr("201 : Another DR command is running (pid %s)" %
                   another_pid)
        else:
            LogDebug(e)
            LogErr('202 : Can not confirm DR command status')
        return 1
    except Exception as e:
        LogDebug(e)
        LogErr('203 : Can not confirm DR command status')
        return 1

    pid = os.getpid()
    try:
        os.write(fd, "%d\n" % pid)
        os.close(fd)
    except Exception as e:
        LogDebug(e)
        LogErr('204 : Can not confirm DR command status')
        return 1

    return 0


def CheckEnvironment():
    '''
    Check system environment before start.
    Check list:
       - are you root user?
       - another dr command is running?
    Return value(int):
       - the condition meets requirements -> 0
       - the condition does not meet requirement(s) -> 1
    '''

    uid = os.getuid()
    if uid:
        LogErr('205 : Must be root for dr operations')
        return 1

    return CheckConcurrentRunning()


def CleanupPIDFile():
    '''
    Cleanup PID File for next run
    Return value: None
    '''
    try:
        os.unlink(DP_PID_FILE)
    except Exception as e:
        LogDebug(e)


def CheckDistroVersion():
    '''
    Check Distribution version number of running system
    Return Value(tuple (str, str)): major, minor version numbers or
    None if error.
    '''
    major, minor = (None, None)

    if os.path.isfile(OS_RELEASE_FILE):
        f = open(OS_RELEASE_FILE, 'r')
        for line in f:
            if '=' in line:
                l, r = line.split('=')
                if l == 'VERSION_ID':
                    tmp = r.rstrip().strip('"').split('.')
                    if len(tmp) == 1:
                        major = tmp[0]
                        minor = '0'
                    elif len(tmp) == 2:
                        major, minor = tmp
                    else:
                        pass
            else:
                pass
        f.close()
        if major is None or minor is None:
            LogDebug("Distribution is unknown")
    else:
        LogDebug("OS release file doesn't exist")

    return (major, minor)

def PrintVersion():
    return "dr command version %s" % DP_CMD_VERSION
