#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# library for IPMI command sending
#
# Copyright(C) 2012 MUNEDA Takahiro
# Copyright(C) 2012 FUJITSU LIMITED
#

import os
import re
import subprocess

import dp

DP_IPMI_LIB_VERSION = '1.0'
IPMI_COMMON_HEADER = ['0x2E', '0xFC', '0xD3', '0x00', '0x00']
SEND_IPMI_BUS_INFO_START = 0x00e0
SEND_IPMI_BUS_INFO_END = 0x00ef
DP_EVENTS = ['SB_ADD', 'SB_REMOVE', 'IOU_ADD', 'IOU_REMOVE']

mac_addr_regexp = re.compile(r'([0-9a-f]{2}:){5}[0-9a-f]{2}$', re.IGNORECASE)


def _SendReady(ready):
    '''
    Send OS ready/not ready by IPMI to notify OS is ready for DP operations.
    @ready(int): OS ready -> 1, OS not ready -> 0
    Return value(int): success -> 0, error -> 1
    '''
    path = "%s/cmdline" % dp.DP_PROCFS
    if os.path.exists(path):
        cmdline, error = dp.ReadFile(path)
        if error:
            dp.LogDebug("%s: Cannot read with %d" % (path, error))
            return 1
        if 'movable_node' in cmdline:
            dp.LogDebug("movable_node must not be set on SLES for DR")
            return 1

    if not isinstance(ready, int) or not 0 <= ready <= 1:
        if isinstance(ready, tuple):
            ready = (ready,)
        dp.LogDebug("%s: wrong ready was passed" % ready)
        return 1

    # Build an IPMI command
    ipmi_command = BuildOSReady(ready)
    if not ipmi_command:
        return 1

    expected_return_code = ['D3', '00', '00']
    response, retval = SendIPMI(ipmi_command)
    if not retval == 0:
        return 1
    if not response == expected_return_code:
        return 1
    return 0


def SendOSReady():
    '''
    Wrapper method for Sending OS Ready notification
    Return value: return value from _SendReady()
    '''
    return _SendReady(1)


def SendOSNotReady():
    '''
    Wrapper method for Sending OS Ready notification
    Return value: return value from _SendReady()
    '''
    return _SendReady(0)


def _SendCompletion(event, code):
    '''
    Send a notification of SB/IOU hotplug completion by IPMI
    @event(str): {SB,IOU}_{ADD,REMOVE}
    @code(int): Error/Progress Code of DP,
    progress -> 0x00e[0f] or 0x00ff, error -> -0x00[12345679abcd]0
    Return value(int): success -> 0, error -> 1
    '''
    if not isinstance(event, str) or not event in DP_EVENTS:
        if isinstance(event, tuple):
            event = (event,)
        dp.LogDebug("%s: wrong event was passed" % event)
        return 1
    if not isinstance(code, int):
        if isinstance(code, tuple):
            code = (code,)
        dp.LogDebug("%s: wrong status was passed" % code)
        return 1

    # Build an IPMI command
    ipmi_command = BuildCompletion(event, code)
    if not ipmi_command:
        return 1

    expected_return_code = ['D3', '00', '00']
    response, retval = SendIPMI(ipmi_command)
    if not retval == 0:
        return 1
    if not response == expected_return_code:
        return 1
    return 0


def SendSBAddCompletion(status):
    '''
    Wrapper method for Sending Completion notification
    @status(int): success -> 0, error -> 0x2, 0x3, 0x7, 0x9, or 0xa
    Return value: success -> 0, error -> 1
    '''
    SB_ADD_STATUS = [0x0, 0x2, 0x3, 0x7, 0x9, 0xa]
    if not isinstance(status, int) or status not in SB_ADD_STATUS:
        if isinstance(status, tuple):
            status = (status,)
        dp.LogDebug("%s: wrong status was passed" % status)
        return 1

    code = None
    if status == 0x0:
        code = 0x00ff
    else:
        code = status * -(0x0010)

    if _SendCompletion('SB_ADD', code):
        return 1
    else:
        return 0


def SendSBRemoveCompletion(status):
    '''
    Wrapper method for Sending Completion notification
    @status(int): success -> 0, error -> 0x1, 0x2, 0x4, 0x7, 0x9, or 0xb
    Return value: success -> 0, error -> 1
    '''
    SB_REMOVE_STATUS = [0x0, 0x1, 0x2, 0x4, 0x7, 0x9, 0xb]
    if not isinstance(status, int) or status not in SB_REMOVE_STATUS:
        if isinstance(status, tuple):
            status = (status,)
        dp.LogDebug("%s: wrong status was passed" % status)
        return 1

    code = None
    if status == 0x0:
        code = 0x00ff
    else:
        code = status * -(0x0010)

    if _SendCompletion('SB_REMOVE', code):
        return 1
    else:
        return 0


def SendIOUAddCompletion(status):
    '''
    Wrapper method for Sending Completion notification
    @status(int): success -> 0, error -> 0x1, 0x2, 0x5, 0x7, or 0xc
    Return value: success -> 0, error -> 1
    '''
    IOU_ADD_STATUS = [0x0, 0x1, 0x2, 0x5, 0x7, 0xc]
    if not isinstance(status, int) or status not in IOU_ADD_STATUS:
        if isinstance(status, tuple):
            status = (status,)
        dp.LogDebug("%s: wrong status was passed" % status)
        return 1

    if status in [0x0, 0x1, 0x2, 0x7]:
        # Do NOT send Progress Code and minor errors in IOU hotplug
        return 0

    code = status * -(0x0010)
    if _SendCompletion('IOU_ADD', code):
        return 1
    else:
        return 0


def SendIOURemoveCompletion(status):
    '''
    Wrapper method for Sending Completion notification
    @status(int): success -> 0, error -> 0x1, 0x2, 0x6, 0x7, or 0xd
    Return value: success -> 0, error -> 1
    '''
    IOU_REMOVE_STATUS = [0x0, 0x1, 0x2, 0x6, 0x7, 0xd]
    if not isinstance(status, int) or status not in IOU_REMOVE_STATUS:
        if isinstance(status, tuple):
            status = (status,)
        dp.LogDebug("%s: wrong status was passed" % status)
        return 1

    if status in [0x0, 0x1, 0x2, 0x7]:
        # Do NOT send Progress Code and minor errors in IOU hotplug
        return 0

    code = status * -(0x0010)
    if _SendCompletion('IOU_REMOVE', code):
        return 1
    else:
        return 0


def SendBusTreeStart(event):
    '''
    Send start command to notify Bus Tree info will be send
    @event(str): IOU_{ADD,REMOVE}
    Return value(int): success -> 0, error -> 1
    Note: This is a part of IOU hotplug and will be called before
    SendIOU{Add,Remove}Completion().
    '''
    return _SendCompletion(event, SEND_IPMI_BUS_INFO_START)


def SendBusTreeEnd(event):
    '''
    Send end command to notify Bus Tree info was sent
    @event(str): IOU_{ADD,REMOVE}
    Return value(int): success -> 0, error -> 1
    Note: This is a part of IOU hotplug and will be called before
    SendIOU{Add,Remove}Completion().
    '''
    return _SendCompletion(event, SEND_IPMI_BUS_INFO_END)


def _SendBusTreeInfo(pci_addr):
    '''
    Send Bus Tree information
    @pci_addr(str): PCI address of PCI bridge
    Return value(int): success -> 0, error -> 1
    '''
    if not isinstance(pci_addr, str) or not dp.pci_addr_regexp.match(pci_addr):
        if isinstance(pci_addr, tuple):
            pci_addr = (pci_addr,)
        dp.LogDebug("%s: wrong pci_addr was passed" % pci_addr)
        return 1

    # Build an IPMI command
    ipmi_command = BuildSetBusTree(pci_addr)
    if not ipmi_command:
        return 1

    expected_return_code = ['D3', '00', '00']
    response, retval = SendIPMI(ipmi_command)
    if not retval == 0:
        return 1
    if not response == expected_return_code:
        return 1
    return 0


def SendBusTrees(pci_addrs):
    '''
    Send Bus Tree information on system
    @pci_addrs(str): list of PCI address
    Return value(int): success -> 0, error-> 1
    Note: This method will not send start/end IPMI commands.  Need to
    send them before and after call this regardless the result of this
    method.  Or someone may keep waiting end commands.
    '''
    if not pci_addrs or not isinstance(pci_addrs, list):
        if isinstance(pci_addrs, tuple):
            pci_addrs = (pci_addrs,)
        dp.LogDebug("%s: wrong pci_addrs was passed" % pci_addrs)
        return 1

    bridges = []
    for pci_addr in pci_addrs:
        if not _IsBridge(pci_addr):
            continue
        dp.LogDebug("PCI Bridge found: %s" % pci_addr)
        bridges.append(pci_addr)

    if not bridges:
        dp.LogDebug('No PCI bridge.')
        return 0
    # Need to sort in PCI address ascending order
    bridges.sort()

    retval = 0
    for pci_addr in bridges:
        retval += _SendBusTreeInfo(pci_addr)
    if retval == 0:
        return 0
    else:
        return 1


def _SendMacAddress(pci_addr):
    '''
    Send MAC address of on board NIC
    @pci_addr(str): PCI address of on board device
    Return value(int): success -> 0, error -> 1
    '''
    if not isinstance(pci_addr, str) or not dp.pci_addr_regexp.match(pci_addr):
        if isinstance(pci_addr, tuple):
            pci_addr = (pci_addr,)
        dp.LogDebug("%s: wrong pci_addr was passed" % pci_addr)
        return 1

    # looking for a file named address which has mac_address
    net_path = "%s/bus/pci/devices/%s/net" % (dp.DP_SYSFS, pci_addr)
    if not os.path.isdir(net_path):
        dp.LogDebug("%s: not a directory." % net_path)
        return 1
    eth_path = None
    for entry in os.listdir(net_path):
        if ('eth' in entry or
            re.match(r'p\d+p\d+', entry) or
            re.match(r'enp\d+s\d+', entry) or
            re.match(r'eno\d', entry) or
            re.match(r'ens\d', entry) or
            re.match(r'em\d', entry)):
            eth_path = "%s/%s" % (net_path, entry)
            break
    mac_addr_path = "%s/address" % eth_path

    mac_addr, errno = dp.ReadFile(mac_addr_path)
    dp.LogDebug("%s has MAC address: %s" % (pci_addr, mac_addr))
    if mac_addr is None:
        return 1
    elif not mac_addr_regexp.match(mac_addr):
        return 1

    # Build an IPMI command
    ipmi_command = BuildSetMacAddress(pci_addr, mac_addr)
    if not ipmi_command:
        return 1

    expected_return_code = ['D3', '00', '00']
    response, retval = SendIPMI(ipmi_command)
    if not retval == 0:
        return 1
    if not response == expected_return_code:
        return 1
    return 0


def SendMacAddresses(pci_addrs):
    '''
    Send MAC address of on board NICs
    @pci_addrs(str): list of PCI address
    Return value(int): success -> 0, error -> 1
    Note: This method will check the pci address in the pci_addrs is
    for an on-board NIC, so no need to check in advance.
    '''
    if not pci_addrs or not isinstance(pci_addrs, list):
        if isinstance(pci_addrs, tuple):
            pci_addrs = (pci_addrs,)
        dp.LogDebug("%s: wrong pci_addrs was passed" % pci_addrs)
        return 1

    onboard_devs = []
    for pci_addr in pci_addrs:
        if not dp.pci_addr_regexp.match("%s" % pci_addr):
            continue
        if _IsOnBoard(pci_addr):
            onboard_devs.append(pci_addr)
    if not onboard_devs:
        dp.LogDebug('No on-board NIC')
        return 0

    retval = 0
    for onboard_dev in onboard_devs:
        retval += _SendMacAddress(onboard_dev)
    if retval == 0:
        return 0
    else:
        return 1


def _CallSetpci(pci_addr, optarg=''):
    '''
    Call setpci command to get PCI device info
    @pci_addr(str): PCI address for target device in seg:bus:dev.fn format
    @optarg(str): optional argument for setpci command
    Return value(str): output of setpci command or None if error happens
    '''
    import shlex

    if not isinstance(pci_addr, str) or not dp.pci_addr_regexp.match(pci_addr):
        if isinstance(pci_addr, tuple):
            pci_addr = (pci_addr,)
        dp.LogDebug("%s: wrong pci_addr was passed" % pci_addr)
        return
    if optarg and not isinstance(optarg, str):
        if isinstance(optarg, tuple):
            optarg = (optarg,)
        dp.LogDebug("%s: wrong optarg was passed" % optarg)
        return

    setpci_path = '/sbin/setpci'
    if dp.IsNonExecutable(setpci_path):
        return

    args = [setpci_path, '-s', pci_addr]
    if optarg:
        opt_list = shlex.split(optarg)
        args += opt_list
    dp.LogDebug("Calling setpci: %s" % (' ').join(args))
    try:
        process = subprocess.Popen(args, stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE)
        process.wait()
    except Exception as e:
        dp.LogDebug(e)
        return
    else:
        out_msgs, err_msgs = process.communicate()
        if process.returncode == 0:
            return out_msgs.rstrip('\n\r')
        else:
            dp.LogDebug("setpci returns with returncode: %d" %
                        process.returncode)
            if out_msgs:
                for out_msg in out_msgs.splitlines():
                    dp.LogDebug("%s" % out_msg)
            if err_msgs:
                for err_msg in err_msgs.splitlines():
                    dp.LogDebug("%s" % err_msg)
            return


def _GetClassProg(pci_addr):
    '''
    Get the value of lower 1 byte from Class Code register
    (offset 09h) by calling setpci command.
    @pci_addr(str): PCI address for target device in seg:bus:dev.fn format
    Return value: output of setpci command.
    '''
    return _CallSetpci(pci_addr, 'CLASS_PROG.b')


def _GetClassCode(pci_addr):
    '''
    Get the value of upper 2 bytes from Class Code register
    (offset 09h) by calling setpci command.
    @pci_addr(str): PCI address for target device in seg:bus:dev.fn format
    Return value: output of setpci command.
    '''
    return _CallSetpci(pci_addr, 'CLASS_DEVICE.w')


def _GetHeaderType(pci_addr):
    '''
    Get the value of Header Type Register (offset 0eh) by calling
    setpci command.
    @pci_addr(str): PCI address for target device in seg:bus:dev.fn format
    Return value: output of setpci command.
    '''
    return _CallSetpci(pci_addr, 'HEADER_TYPE.b')


def _GetBaseAddress2(pci_addr):
    '''
    Get the value of Base Address 2 Register (offset 18h) by calling
    setpci command.
    @pci_addr(str): PCI address for target device in seg:bus:dev.fn format
    Return value: output of setpci command.
    '''
    return _CallSetpci(pci_addr, 'BASE_ADDRESS_2.l')


def _GetPciExpCapRegister(pci_addr):
    '''
    Get the value of PCI Express Capabilities Register
    (offset CAP + 10h) by calling setpci command
    @pci_addr(str): PCI address for target device in seg:bus:dev.fn format
    Return value: output of setpci command
    '''
    return _CallSetpci(pci_addr, 'CAP_EXP.l')


def _IsBridge(pci_addr):
    '''
    Check a PCI device is a bridge or not
    @pci_addr(str): PCI address for target device in seg:bus:dev.fn format
    Return value: a bridge -> 1, not a bridge or error -> 0
    '''
    if not isinstance(pci_addr, str) or not dp.pci_addr_regexp.match(pci_addr):
        if isinstance(pci_addr, tuple):
            pci_addr = (pci_addr,)
        dp.LogDebug("%s: wrong pci_addr was passed" % pci_addr)
        return 0

    output = _GetClassCode(pci_addr)
    dp.LogDebug("ClassCode: %s" % output)
    if not output:
        return 0
    if not re.match(r'[0-9a-f]{4}$', output):
        return 0
    class_device = output

    base_class = class_device[0:2]
    if not base_class == '06':
        return 0

    output = _GetHeaderType(pci_addr)
    dp.LogDebug("HeaderType: %s" % output)
    if not output:
        return 0
    if not re.match(r'[0-9a-f]{2}$', output):
        return 0
    if not output == '01' and not output == '81':  # single- or multi-function
        return 0

    return 1


def _IsOnBoard(pci_addr):
    '''
    Check a PCI device is on-board or not(on hotplug slot)
    @pci_addr(str): PCI address for target device in seg:bus:dev.fn format
    Return value(int): on-board -> 1, not on board -> 0, error -> -1
    '''
    REG_SLOT_IMPLEMENTED = 0x01000000

    if not isinstance(pci_addr, str) or not dp.pci_addr_regexp.match(pci_addr):
        if isinstance(pci_addr, tuple):
            pci_addr = (pci_addr,)
        dp.LogDebug("%s: wrong pci_addr was passed" % pci_addr)
        return -1

    path = "%s/bus/pci/devices/%s" % (dp.DP_SYSFS, pci_addr)
    try:
        realpath = os.path.realpath(path)
        parent_addr = realpath.split('/')[-2]
    except Exception as e:
        dp.LogDebug(e)
        return -1

    cap_exp_register = _GetPciExpCapRegister(parent_addr)
    if not cap_exp_register:
        return -1
    if int("%s" % cap_exp_register, 16) & REG_SLOT_IMPLEMENTED:
        return 0
    else:
        return 1


def BuildCompletion(event, code):
    '''
    Build an IPMI command for 'Error/Progress Code of DP'.
    @event(str): {SB,IOU}_{ADD,REMOVE}
    @code(int): progress -> code >= 0, error -> code < 0
    Return value(str): IPMI command or None if error happens
    '''
    COMPLETION_LENGTH = 5 + 5  # (IPMI_COMMON_HEADER) + (data)

    if not isinstance(event, str) or not event in DP_EVENTS:
        if isinstance(event, tuple):
            event = (event,)
        dp.LogDebug("%s: wrong event was passed" % event)
        return None
    if not isinstance(code, int) or not 0x0010 <= abs(code) <= 0x00ff:
        if isinstance(code, tuple):
            code = (code,)
        dp.LogDebug("%s: wrong code was passed" % code)
        return None

    if code >= 0:
        event_type = ['0x02']  # Will send Progress Code of DP
    else:
        event_type = ['0x01']  # Will send Error Code of DP

    if event == 'SB_ADD':
        event_id = '0x00'
    elif event == 'SB_REMOVE':
        event_id = '0x01'
    elif event == 'IOU_ADD':
        event_id = '0x02'
    else:  # event == 'IOU_REMOVE'
        event_id = '0x03'

    # event_code(2 bytes) should be sent in LSB First.
    # The range of 'abs(code)' is from 0x10 to 0xff.  To swap
    # upper/lower bytes, simply add '00' behind
    event_code = "%02x00" % abs(code)
    event_status = [event_id,
                    "0x%s" % event_code[0:2], "0x%s" % event_code[2:4]]

    completion_header = IPMI_COMMON_HEADER + ['0x07']
    ipmi_command_list = completion_header + event_type + event_status
    ipmi_command = (' ').join(ipmi_command_list)
    if not len(ipmi_command_list) == COMPLETION_LENGTH:
        dp.LogDebug("Incorrect ipmi_command: %s" % ipmi_command)
        return

    dp.LogDebug("ipmi_command: %s" % ipmi_command)
    return ipmi_command


def BuildOSReady(ready):
    '''
    Build an IPMI command for 'OS ready'.
    @ready(int): 0 -> Not Ready, 1 -> Ready
    Return value(str): IPMI command or None if error happens
    '''
    OS_READY_LENGTH = 5 + 3  # (IPMI_COMMON_HEADER) + (data)
    NOT_READY = 0
    READY = 1

    if not isinstance(ready, int) or not NOT_READY <= ready <= READY:
        if isinstance(ready, tuple):
            ready = (ready,)
        dp.LogDebug("%s: wrong ready was passed" % ready)
        return

    os_ready_header = IPMI_COMMON_HEADER + ['0x07']
    event_type = ['0x00']
    event_status = ["0x%02x" % ready]

    ipmi_command_list = os_ready_header + event_type + event_status
    ipmi_command = (' ').join(ipmi_command_list)
    if not len(ipmi_command_list) == OS_READY_LENGTH:
        dp.LogDebug("Incorrect ipmi_command: %s" % ipmi_command)
        return

    dp.LogDebug("ipmi_command: %s" % ipmi_command)
    return ipmi_command


def BuildSetBusTree(pci_addr):
    '''
    Build an IPMI command for 'Set Bus Tree'
    @pci_addr(str): PCI address of PCI bridge
    Return value(str): IPMI command or None if error happens
    '''
    SET_BUS_TREE_LENGTH = 5 + 9  # (IPMI_COMMON_HEADER) + (data)

    if not isinstance(pci_addr, str) or not dp.pci_addr_regexp.match(pci_addr):
        if isinstance(pci_addr, tuple):
            pci_addr = (pci_addr,)
        dp.LogDebug("%s: wrong pci_addr was passed" % pci_addr)
        return None

    set_bus_tree_header = IPMI_COMMON_HEADER + ['0x01']

    output = _GetClassProg(pci_addr)
    if not output:
        return None
    if not re.match(r'[0-9a-f]{2}$', output):
        return None
    reg_level_prog_if = ["0x%s" % output]

    output = _GetClassCode(pci_addr)
    if not output:
        return None
    if not re.match(r'[0-9a-f]{4}$', output):
        return None
    sub_base_class = ["0x%s" % output[2:4], "0x%s" % output[0:2]]

    pci_addr_list = pci_addr.split(':')
    dev, fn = pci_addr_list[2].split('.')
    devfn = (int(dev, 16) << 3) + int(fn, 16)
    seg = int(pci_addr_list[0], 16)
    devfn_seg = ["0x%02x" % devfn, "0x%02x" % seg]

    output = _GetBaseAddress2(pci_addr)
    if not output:
        return None
    if not re.match(r'[0-9a-f]{8}$', output):
        return None
    bus_sec_sub_bus = ["0x%s" % output[6:8], "0x%s" % output[4:6],
                       "0x%s" % output[2:4]]

    ipmi_command_list = (set_bus_tree_header + reg_level_prog_if
                         + sub_base_class + devfn_seg + bus_sec_sub_bus)
    ipmi_command = (' ').join(ipmi_command_list)
    if not len(ipmi_command_list) == SET_BUS_TREE_LENGTH:
        dp.LogDebug("Incorrect ipmi_command: %s" % ipmi_command)
        return None

    dp.LogDebug("ipmi_command: %s" % ipmi_command)
    return ipmi_command


def BuildSetMacAddress(pci_addr, mac_addr):
    '''
    Build an IPMI command for 'Set MAC Address'
    @pci_addr(str): PCI address for target device in seg:bus:dev.fn format
    @mac_addr(str): MAC address of this device in 01:23:45:67:89:ab format
    Return value(str): IPMI command or None if error happens
    '''
    SET_MAC_ADDRESS_LENGTH = 5 + 12  # (IPMI_COMMON_HEADER) + (data)

    if not isinstance(pci_addr, str) or not dp.pci_addr_regexp.match(pci_addr):
        if isinstance(pci_addr, tuple):
            pci_addr = (pci_addr,)
        dp.LogDebug("%s: wrong pci_addr was passed" % pci_addr)
        return None
    if not isinstance(mac_addr, str) or not mac_addr_regexp.match(mac_addr):
        if isinstance(mac_addr, tuple):
            mac_addr = (mac_addr,)
        dp.LogDebug("%s: wrong mac_addr was passed" % mac_addr)
        return None

    set_mac_addr_header = IPMI_COMMON_HEADER + ['0x04', '0xE0', '0x00']
    ipmi_command = None

    octets = []
    for octet in mac_addr.split(':'):
        octets.append("0x%s" % octet)

    pci_addr_list = pci_addr.split(':')
    seg = int(pci_addr_list[0], 16)
    bus = pci_addr_list[1]
    dev, fn = pci_addr_list[2].split('.')
    devfn = (int(dev, 16) << 3) + int(fn, 16)
    seg_bus_devfn = ["0x%02x" % seg, "0x%s" % bus, "0x%02x" % devfn]

    ipmi_command_list = set_mac_addr_header + octets + seg_bus_devfn
    ipmi_command = (' ').join(ipmi_command_list)
    if not len(ipmi_command_list) == SET_MAC_ADDRESS_LENGTH:
        dp.LogDebug("Incorrect ipmi_command: %s" % ipmi_command)
        return None

    dp.LogDebug("ipmi_command: %s" % ipmi_command)
    return ipmi_command


def SendIPMI(ipmi_command):
    '''
    Send IPMI command to MMB via BMC by calling ipmitools
    @ipmi_command(str): IPMI command
    Return value(tuple (str, int)): output from ipmitools or None and
    return value of ipmitool command or None
    '''
    if not ipmi_command or not isinstance(ipmi_command, str):
        if isinstance(ipmi_command, tuple):
            ipmi_command = (ipmi_command,)
        dp.LogDebug("%s: wrong ipmi_command was passed" % ipmi_command)
        return (None, None)

    ipmitool_path = '/usr/bin/ipmitool'
    if dp.IsNonExecutable(ipmitool_path):
        return (None, None)

    args = [ipmitool_path, '-d', '0', 'raw'] + ipmi_command.split(' ')
    dp.LogDebug("Calling ipmitool: %s" % (' ').join(args))
    try:
        process = subprocess.Popen(args, stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE)
        process.wait()
    except Exception as e:
        dp.LogDebug(e)
        return (None, None)
    else:
        out_msgs, err_msgs = process.communicate()
        dp.LogDebug("ipmitool returns with returncode: %d" %
                    process.returncode)
        if out_msgs:
            for out_msg in out_msgs.splitlines():
                dp.LogDebug("%s" % out_msg)
        if err_msgs:
            for err_msg in err_msgs.splitlines():
                dp.LogDebug("%s" % err_msg)

        # convert str into list for easy comparison, and remove a
        # blank in line head which is added by ipmitool
        response = str(out_msgs).rstrip('\n\r').upper().split(' ')[1:]
        return (response, process.returncode)
