#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Enable non-Home SB(s) as removable SB
#
# Copyright(C) 2013 MUNEDA Takahiro
# Copyright(C) 2013 FUJITSU LIMITED
#

import os
import re
import sys

import dp

mem_entry_regexp = re.compile(r'^memory([\d]+)$')
node_entry_regexp = re.compile(r'node([\d]+)$')


def _LNXCPUToNode(cpu_path):
    '''
    Check which NUMA node this CPU is on
    @cpu_path(str): absolute path of the LNXCPU:NN or ACPI0007:NN
    Return value(str): NUMA node number or None if error
    '''
    path = "%s/physical_node" % cpu_path
    if not os.path.isdir(path):
        dp.LogDebug("%s: not a directory" % path)
        return None
    for node_entry in os.listdir(path):
        if not node_entry_regexp.match(node_entry):
            continue
        return node_entry_regexp.sub(r'\1', node_entry)


def _DetectNodesOnHomeSB():
    '''
    Detect all NUMA node numbers on Home SB
    Return value(list of str): NUMA node numbers, or None if not
    exists or error
    '''
    path = "%s/bus/acpi/devices/ACPI0004:00" % dp.DP_SYSFS
    if not os.path.isdir(path):
        dp.LogDebug("%s: not a directry" % path)
        return None

    nodes = []
    for entry in os.listdir(path):
        if not 'LNXCPU:' in entry and not 'ACPI0007' in entry:
            continue
        cpu_path = "%s/%s" % (path, entry)
        node_num = _LNXCPUToNode(cpu_path)
        # On RHEL7.1 with disabled HT, LNXCPU:XX/ACPI0007:XX for HT is not
        # linked node. So the if _LNXCPUToNode returns None, continue it.
        if node_num is None:
            major, minor = dp.CheckDistroVersion()
            if major == '7' and minor == '0':
                dp.LogDebug("%s: does not have correct node info" % entry)
                return None
            continue
        if node_num in nodes:
            continue
        nodes.append(node_num)

    if len(nodes) is 0:
        dp.LogDebug("Home SB does not have node info")
        return None    

    dp.LogDebug("%s: list of nodes on Home SB" %
                dp.ListToAbbrev(nodes)[0])
    return nodes


def _GetMemoryEntryList(node_num, reverse=False):
    '''
    Get a list of memory entry under this node as a list
    @node_num(str): NUMA node number
    @reverse(int): A switch to turn on reverse
    Return value(str): List of memory entry or None if not found or
    error
    '''
    path = "%s/devices/system/node/node%s" % (dp.DP_SYSFS, node_num)

    mem_num_list = []
    for mem_entry in os.listdir(path):
        if not mem_entry_regexp.match(mem_entry):
            continue
        mem_num = mem_entry_regexp.sub(r'\1', mem_entry)
        mem_num_list.append(mem_num)

    if len(mem_num_list) == 0:
        dp.LogDebug('No memory available on this node, really?')
        return None

    mem_num_list.sort(lambda x, y: cmp(int(x), int(y)), reverse=reverse)

    mem_list = []
    for mem_num in mem_num_list:
        mem = "memory%s" % mem_num
        mem_list.append(mem)
    return mem_list


def _EnableMemoryOnNode(node_num):
    '''
    Online all offlined memory on this NUMA node as removable
    @node_num(str): NUMA node number
    Return value(int): success -> 0, error -> 1
    '''
    path = "%s/devices/system/node/node%s" % (dp.DP_SYSFS, node_num)
    if not os.path.isdir(path):
        dp.LogDebug("%s: not a directory" % path)
        return 1

    # Movable zone must reside after Normal zone.  To ensure this
    # order, Kernel reject to online a memory as online_movable if
    # there is an offlined a memory after the memory.
    # So, need to online a memory in upper section first
    mem_entry_list = _GetMemoryEntryList(node_num, reverse=True)
    if mem_entry_list is None:
        return 1

    for mem_entry in mem_entry_list:
        if not mem_entry_regexp.match(mem_entry):
            continue
        power_path = "%s/%s/state" % (path, mem_entry)
        if dp.WriteFile(power_path, 'online_movable'):
            dp.LogDebug("%s: failed to online_movable" % mem_entry)
            return 1
    return 0


def _EnableMemoryOnNonHomeSB():
    '''
    Online all offlined memory on the system as removable
    Return value(int): success -> 0, error -> 1

    Note: the reason not to check all memory device status under
    /sys/devices/system/memory is amount of those entries.  If the
    system has 12TiB of memory, the number of sysfs entries become
    more than 100k.  Therefore, we check the memory entry on non Home
    SBs only.
    '''
    nodes = _DetectNodesOnHomeSB()
    if nodes is None:
        dp.LogDebug('Could not detect node num on Home SB')
        return 1

    path = "%s/devices/system/node" % dp.DP_SYSFS
    if not os.path.isdir(path):
        dp.LogDebug("%s: not a directory" % path)
        return 1

    for node_entry in os.listdir(path):
        if not node_entry_regexp.match(node_entry):
            continue
        node_num = node_entry_regexp.sub(r'\1', node_entry)
        if node_num in nodes:
            continue
        if _EnableMemoryOnNode(node_num):
            return 1
    return 0


def _FixNodeNumMismatchOnSB(sb_dir):
    '''
    Fix the symlink to NUMA node from CPU on this SB
    @sb_dir(str): a directory correspond to the SB which has CPUs
    which have incorrect symlinks to the NUMA node (ex. ACPI0004:01)
    Return value(int): success -> 0, error -> 1
    Note: The symlink locates under
    /sys/bus/acpi/devices/ACPI0004:XX/{LNXCPU:YY or ACPI0007}/physical_node
    and named nodeZ.  This symlink will be fixed by kernel when the CPU is
    onlined.
    '''
    sb_path = "%s/bus/acpi/devices/%s" % (dp.DP_SYSFS, sb_dir)
    if not os.path.isdir(sb_path):
        dp.LogDebug("%s: not a directory" % sb_path)
        return 1

    power_path_list = []
    for entry in os.listdir(sb_path):
        if not 'LNXCPU:' in entry and not 'ACPI0007' in entry:
            continue

        power_path = "%s/%s/physical_node/online" % (sb_path, entry)
        # On RHEL7.1 with disabled HT, LNXCPU:XX/ACPI0007 for HT does not
        # have online file. So the if condition is true, skip it.
        if not os.path.isfile(power_path):
            continue
        power_path_list.append(power_path)

    if len(power_path_list) == 0:
        return 0

    # All CPUs on a same CPU socket need to belong to a same NUMA
    # node, or Kernel shows warnigns while onlining the CPU.
    # So, power off all CPUs, then re-power on all of them
    retval = 0
    for power_path in power_path_list:
        if dp.WriteFile(power_path, dp.POWEROFF):
            dp.LogDebug("%s: failed to offline" % power_path)
            retval += 1
    if retval:
        return 1

    for power_path in power_path_list:
        if dp.WriteFile(power_path, dp.POWERON):
            dp.LogDebug("%s: failed to online" % power_path)
            retval += 1
    if retval:
        return 1
    else:
        return 0


def _FixNodeNumMismatch():
    '''
    Fix the symlink to NUMA node from CPU on Non-HomeSB
    Return value(int): success -> 0, error -> 1
    '''
    path = "%s/bus/acpi/devices" % dp.DP_SYSFS
    if not os.path.isdir(path):
        dp.LogDebug("%s: not a directory" % path)
        return 1

    for entry in os.listdir(path):
        if not 'ACPI0004:' in entry:
            continue
        if entry == 'ACPI0004:00':
            continue
        # On RHEL7.1 or later, ACPI0004 directory always exists.
        # To check ACPI0004 is poweron or not, the status file must
        # be checked on RHEL7.1 or later
        sta_path = "%s/%s/status" % (path, entry)
        if os.path.exists(sta_path):
            sta, error = dp.ReadFile(sta_path)
            if sta is None:
                return 1
            if int(sta) == 0:
                continue
        if _FixNodeNumMismatchOnSB(entry):
            return 1
    return 0


def main():
    '''
    Setup devices on non Home SBs to enable future hot-removal
    Return value(int): success -> 0, error -> 1
    '''
    if _EnableMemoryOnNonHomeSB():
        dp.LogDebug('Failed to enable some memory')
        return 1
    if _FixNodeNumMismatch():
        dp.LogDebug('Failed to fix node number mismatch')
        return 1
    return 0


def pre_main():
    '''
    Check system environment before processing main()
    Return value(int): return value from main() or error -> 1
    '''
    uid = os.getuid()
    if uid:
        dp.LogErr('750 : Must be root for dr operations')
        return 1

    retval = main()
    if retval:
        dp.LogErr('751 : Failed to setup non Home SBs')
    return retval


if __name__ == '__main__':
    sys.exit(pre_main())
