#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# DR event daemon
#
# Copyright(C) 2013 MUNEDA Takahiro
# Copyright(C) 2013 FUJITSU LIMITED
#

import os
import re
import sys
import socket

import dp
import notify


def _CalledByExpectedCaller():
    '''
    Check whether this process is called by expected caller
    Return value(int): success -> 0, error -> 1
    '''
    name_regexp = re.compile(r'^Name:\s*([\w]+)\n$')
    expected_name = 'systemd'

    ppid = os.getppid()
    status, error = dp.ReadlinesFile("/proc/%s/status" % ppid)
    if not status:
        dp.LogDebug('Could not get parent name')
        return 1
    name = None
    for line in status:
        if name_regexp.match(line):
            name = name_regexp.sub(r'\1', line)
    if not name:
        dp.LogDebug('Could not get parent name')
        return 1

    if not name == expected_name:
        dp.LogDebug("%s is called by unexpected caller %s" %
                    (sys.argv[0], name))
        return 1
    return 0


def _ParseEvent(event):
    '''
    Parse a event which is sent from DR event notifier
    @event(str): a message which is sent from event notifier
    Return value(tuple (str, str)): a pair of devpath and action or
    None if error happens

    Sample message (format: <DEVPATH>_FJSVDR_<ACTION>):
    /bus/acpi/devices/LNXSYSTM:00/device:00/ACPI0004:00_FJSVDR_online
    '''
    event_regexp = re.compile(r'^([\w:/]+)_FJSVDR_([\w:/]+)$')

    if not event or not isinstance(event, str):
        dp.LogDebug("%s: wrong event was passed" % str(event))
        return (None, None)
    if not event_regexp.match(event):
        dp.LogDebug("%s: wrong event was passed" % event)
        return (None, None)

    devpath = None
    action = None
    try:
        devpath, action = event.split('_FJSVDR_')
    except ValueError:
        dp.LogDebug("%s: does not contain a marker chars '_FJSVDR_'"
                    % event)
        return (None, None)
    return (devpath, action)


def _EvokeDPCommand(devpath, action):
    '''
    Evoke udev_dp_bridge.py to pass over DR operation
    @devpath(str): DEVPATH which is set by udev
    @action(str): ACTION which is set by udev
    Return value(int): success -> 0, parent error -> 1, child error -> 2
    Note: Success means forked child process.  It does not guarantee
    that child process sucessfully evoke main DR command)
    '''
    if not devpath or not isinstance(devpath, str):
        dp.LogDebug("%s: wrong devpath was passed" % str(devpath))
        return 1
    if not action or not isinstance(action, str):
        dp.LogDebug("%s: wrong action was passed" % str(action))
        return 1

    udev_dp_bridge_path = '/opt/FJSVdr-util/libexec/udev_dp_bridge.py'
    if dp.IsNonExecutable(udev_dp_bridge_path):
        return 1
    args = [udev_dp_bridge_path, devpath, action]
    dp.LogDebug("Calling udev_dp_bridge.py: %s %s" %
                (udev_dp_bridge_path, (' ').join(args)))

    try:
        pid = os.fork()
    except OSError as e:
        dp.LogDebug(e)
        return 1
    except Exception as e:
        dp.LogDebug(e)
        return 1
    else:
        if pid > 0:  # parent process
            return 0

    # only child process
    try:
        os.execv(udev_dp_bridge_path, args)
    except Exception as e:
        dp.LogErr(e)
        return 2
    else:
        # Should not reach here
        return 2


def main():
    '''
    A daemon to receive event notifications from DR event notifier which
    is evoked by udev rule
    Return vlaue(int): error -> 1
    '''
    s = notify._OpenDPSocket(server=1)
    if not s:
        dp.LogErr('722 : Could not setup socket')
        return 1

    # Keep waiting notifications until the parent service will be
    # stopped by systemd.  systemd will kill me while stopping the
    # parent service, so do not return 0 intentionally
    while 1:
        dp.LogDebug('Waiting a socket connection from notifier')
        src_socket, address = s.accept()
        dp.LogDebug('Accepted a socket connection')
        src_socket.settimeout(notify.DP_SOCKET_TIMEOUT)

        event = ''
        try:
            event = src_socket.recv(1024)
            dp.LogDebug("%s: received from event notifier as event" %
                        event)
        except Exception as e:
            dp.LogDebug(
                "%s: error happend while waiting an event message" % e)
            dp.LogErr('723 : Internal error, but waits next event')
            continue

        try:
            sent = src_socket.send(event)
            dp.LogDebug("Sent back %d/%d bytes of data" %
                        (sent, len(event)))
        except Exception as e:
            dp.LogDebug("%s: failed to sent confirmation" % e)

        go_no_go = 'No Go'
        try:
            go_no_go = src_socket.recv(1024)
            dp.LogDebug("%s: received from event notifier as go_no_go"
                        % go_no_go)
        except Exception as e:
            dp.LogDebug(
                "%s: error happened while waiting a go_no_go message" % e)
            dp.LogErr('724 : Internal error, but waits next event')
            continue
        if not go_no_go == 'Go':
            dp.LogErr('725 : Internal error, but waits next event')
            continue

        devpath, action = _ParseEvent(event)
        if not action or not devpath:
            dp.LogDebug("%s:%s, devpath or action was not set" %
                        (action, devpath))
            dp.LogErr('726 : Internal error, but waits next event')
            continue

        retval = _EvokeDPCommand(devpath, action)
        if retval == 1:
            # error happend in parent process, but keep running
            dp.LogErr('727 : Internal error, but waits next event')
            continue
        elif retval == 2:
            # error happend in child process and this process is
            # useless, then returns immediately.  Hasta la vista!
            dp.LogErr('728 : Internal error, but waits next event')
            return 1

    # Shoud not reach here
    notify._CloseDPSocket(s, server=1)
    return 1


def pre_main():
    '''
    Check system environment before processing main()
    Return value(int): return value from main() or error -> 1
    '''
    uid = os.getuid()
    if uid:
        dp.LogErr('720 : Must be root for dr operations')
        return 1
    if _CalledByExpectedCaller():
        dp.LogErr('721 : Must be started by systemd.service')
        return 1

    dp.AvoidOom()
    return main()


if __name__ == '__main__':
    sys.exit(pre_main())
