#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Send DR event notifier to DR event daemon
#
# Copyright(C) 2013 MUNEDA Takahiro
# Copyright(C) 2013 FUJITSU LIMITED
#

import os
import sys
import socket

import dp
import notify


def _SendMessage(message, dest_socket):
    '''
    Send a message to the destination and confirm that it is surely
    sent.
    @message(str): A message to be sent
    @dest_socket(socket): A socket object of the destination
    Return value(int): success -> 0, error -> 1
    '''
    if not message or not isinstance(message, str):
        dp.LogDebug("%s: wrong message was passed" % str(message))
        return 1

    if not dest_socket or not isinstance(dest_socket,
                                         socket._socketobject):
        dp.LogDebug("%s: wrong dest_socket was passed" %
                    str(dest_socket))
        return 1

    # Send a message
    try:
        sent = dest_socket.send(message)
    except Exception as e:
        dp.LogDebug(e)
        return 1
    else:
        if not sent == len(message):
            dp.LogDebug("Only sent %d/%d bytes of message" % (sent,
                                                              len(message)))
            return 1

    # Check it was successfully sent
    try:
        dest_socket.settimeout(notify.DP_SOCKET_TIMEOUT)
        received = dest_socket.recv(1024)
    except Exception as e:
        dp.LogDebug(e)
        return 1
    else:
        if not received == message:
            dp.LogDebug("%s: returned message from DR Event Server" %
                        received)
            return 1

    # Send a Go (or No Go?)
    go = 'Go'
    try:
        sent = dest_socket.send(go)
    except Exception as e:
        dp.LogDebug(e)
        return 1
    else:
        if not sent == len(go):
            dp.LogDebug("Only sent %d/%d bytes of message" % (sent,
                                                              len(go)))
            return 1
    return 0


def _NotifyDPEvent(devpath, action):
    '''
    Send a message to daemon through UNIX socket
    @devpath(str): DEVPATH which is set by udev
    @action(str): ACTION which is set by udev
    Return value(int): success -> 0, error -> 1
    '''
    # just check its existence and type, value itself will be checked by dplib
    if not devpath or not isinstance(devpath, str):
        dp.LogDebug("%s: wrong devpath was passed" % str(devpath))
        return 1
    if not action or not isinstance(action, str):
        dp.LogDebug("%s: wrong action was passed" % str(action))
        return 1
    message = "%s_FJSVDR_%s" % (devpath, action)

    s = notify._OpenDPSocket(server=0)
    if not s:
        dp.LogErr('702 : Could not setup socket')
        return 1
    retval = 0
    if _SendMessage(message, s):
        dp.LogErr("703 : %s, %s : Failed to send event" % (devpath,
                                                           action))
        retval = 1
    notify._CloseDPSocket(s, server=0)

    return retval


def main():
    '''
    Notify a DR event to daemon via UNIX socket
    Return value(int): success -> 0, error -> 1
    '''
    uid = os.getuid()
    if uid:
        dp.LogErr('700 : Must be root for dr operations')
        return 1
    if not len(sys.argv) == 3:
        dp.LogErr("701 : %s : Invalid argument" % (' ').join(sys.argv))
        return 1

    devpath = sys.argv[1]
    action = sys.argv[2]  # online or offline

    retval = _NotifyDPEvent(devpath, action)
    return retval


if __name__ == '__main__':
    sys.exit(main())
