#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# dr command for FJSVdr-utils
#
# Copyright(C) 2012 MUNEDA Takahiro
# Copyright(C) 2012 FUJITSU LIMITED
#

# This program is called via udev_dp_bridge.sh

import os
import re
import sys
import string

import dp
import ipmi
sys.path.append(dp.DP_CORE_PATH)


def _GetSBNumFromDevpath(devpath):
    '''
    Get SB number from DEVPATH
    @devpath(str): DEVPATH which is set by udev
    Return value(int): sun number for the SB.  If not found, 'None'
    '''
    if not devpath or not isinstance(devpath, str):
        return None

    sun_path = "%s%s/sun" % (dp.DP_SYSFS, devpath)
    sun, error = dp.ReadFile(sun_path)
    return sun


def _ChangeContainerpath(devpath):
    '''
    Change devpath to /bus/acpi/devices/ACPI0004:XX
    @devpath(str): DEVPATH which is set by udev
    Return value(str): /bus/acpi/devices/ACPI0004:XX

    Devpaths which set by udev changed on RHEL7.1 or later are as
    follows:
      RHEL7.0: /bus/acpi/devices/ACPI0004:XX.
      RHEL7.1 or later: /devices/system/container/ACPI0004:XX
    The function changes to same devpath as RHEL7.0
    '''

    retval = devpath.replace('/devices/system/container', '/bus/acpi/devices')

    return retval

def _Ignoreaction(action):
    '''
    Ignore action
    @action(str): ACTION which is set by udev
    Return value(int): ignore -> True, not ignore -> False

    For RHEL7.1 or later and SLES12, change uevent is sent to udev for notifying
    SB hot remove event. But when hot removing SB on RHEL7.1 or later,
    change and offline uevents are sent. So udev catches them and send
    them to udev_dp_bridge.py. In this case, udev_dp_bridge.py must ignore
    offline uevent.
    And RHEL7.0 or RHEL6 must ignore change uevent because udev rule
    changes
    '''
    return False


def _CalledByExpectedCaller():
    '''
    Check whether this process is called by expected caller
    Return value(int): success -> 1, error -> 0
    '''
    name_regexp = re.compile(r'^Name:\s*([\w]+)\n$')

    ppid = os.getppid()
    status, error = dp.ReadlinesFile("/proc/%s/status" % ppid)
    if not status:
        dp.LogDebug('Could not get parent name')
        return 0
    name = None
    for line in status:
        if name_regexp.match(line):
            name = name_regexp.sub(r'\1', line)
    if not name:
        dp.LogDebug('Could not get parent name')
        return 0

    major = dp.CheckDistroVersion()[0]
    if int(major) >= 12:
        expected_name = 'systemd'
    else:
        expected_name = 'udev_dp_bridge.sh'
    if name == expected_name:
        return 1
    else:
        dp.LogDebug("%s is called by unexpected caller %s" %
                    (sys.argv[0], name))
        return 0


def main():
    '''
    Call main method of dp_{add,rm}.py
    Return value(int): success -> 0, if error happens -> 1
    '''
    if not _CalledByExpectedCaller:
        return 1

    if not len(sys.argv) == 3:
        dp.LogErr("246 : %s: Invalid argument" % (' ').join(sys.argv))
        return 1
    devpath = sys.argv[1]
    action = sys.argv[2]  # online or offline

    devpath = _ChangeContainerpath(devpath)

    if _Ignoreaction(action):
        dp.LogDebug("%s is ignored" % action)
        return 1

    sb_num = _GetSBNumFromDevpath(devpath)
    if not sb_num:
        if action == 'online':
            major, minor = dp.CheckDistroVersion()
            if major == '6' and minor == '4':
                # RHEL6.4 does not provide _SUN for SB via sysfs
                sb_dev = None
            else:
                dp.LogDebug("%s.%s: sun entry must exist on this system" %
                            (major, minor))
                dp.LogErr('484 : Internal Error')
                return 1
        else:
            dp.LogErr("247 : %s: Invalid argument" % devpath)
            return 1
    else:
        sb_dev = "SB%s" % sb_num

    if action == 'online':
        subcommand = 'add'
        dp.LogInfo('800 : Detected SB hot-add')
    elif action == 'offline':
        subcommand = 'rm'
        dp.LogInfo('820 : Detected SB hot-remove')
    elif action == 'change':
        subcommand = 'rm'
        dp.LogInfo('820 : Detected SB hot-remove')
    else:
        dp.LogErr("248 : %s: Invalid argument" % action)
        return 1

    dp.AvoidOom()
    if dp.CheckEnvironment():
        if subcommand == 'add':
            ipmi.SendSBAddCompletion(0x9)
        else:
            ipmi.SendSBRemoveCompletion(0x9)
        return 1

    if subcommand == 'add':
        import dp_add
        if sb_dev:
            # provide SB number
            retval = dp_add.main(sb_dev)
        else:
            # no SB number, so provide devpath instead on RHEL6.4
            retval = dp_add.main(devpath)
    else:
        import dp_rm
        retval = dp_rm.main(sb_dev)

    dp.LogDebug('All process has finished, remove PID file')
    dp.CleanupPIDFile()
    return retval

if __name__ == '__main__':
    sys.exit(main())
